# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("api-ms-win-core-file-fromapp-l1-1-0.dll")
prototypes = \
    {
        # 
        'CopyFileFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=True, label="Int32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpExistingFileName", "lpNewFileName", "bFailIfExists"]),
        # 
        'CreateDirectoryFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimStruct({"nLength": SimTypeInt(signed=False, label="UInt32"), "lpSecurityDescriptor": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "bInheritHandle": SimTypeInt(signed=True, label="Int32")}, name="SECURITY_ATTRIBUTES", pack=False, align=None), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpPathName", "lpSecurityAttributes"]),
        # 
        'CreateFileFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"nLength": SimTypeInt(signed=False, label="UInt32"), "lpSecurityDescriptor": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "bInheritHandle": SimTypeInt(signed=True, label="Int32")}, name="SECURITY_ATTRIBUTES", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["lpFileName", "dwDesiredAccess", "dwShareMode", "lpSecurityAttributes", "dwCreationDisposition", "dwFlagsAndAttributes", "hTemplateFile"]),
        # 
        'CreateFile2FromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "dwFileAttributes": SimTypeInt(signed=False, label="UInt32"), "dwFileFlags": SimTypeInt(signed=False, label="UInt32"), "dwSecurityQosFlags": SimTypeInt(signed=False, label="UInt32"), "lpSecurityAttributes": SimTypePointer(SimStruct({"nLength": SimTypeInt(signed=False, label="UInt32"), "lpSecurityDescriptor": SimTypePointer(SimTypeBottom(label="Void"), offset=0), "bInheritHandle": SimTypeInt(signed=True, label="Int32")}, name="SECURITY_ATTRIBUTES", pack=False, align=None), offset=0), "hTemplateFile": SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)}, name="CREATEFILE2_EXTENDED_PARAMETERS", pack=False, align=None), offset=0)], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["lpFileName", "dwDesiredAccess", "dwShareMode", "dwCreationDisposition", "pCreateExParams"]),
        # 
        'DeleteFileFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpFileName"]),
        # 
        'FindFirstFileExFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="FINDEX_INFO_LEVELS"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="FINDEX_SEARCH_OPS"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), arg_names=["lpFileName", "fInfoLevelId", "lpFindFileData", "fSearchOp", "lpSearchFilter", "dwAdditionalFlags"]),
        # 
        'GetFileAttributesExFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="GET_FILEEX_INFO_LEVELS"), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpFileName", "fInfoLevelId", "lpFileInformation"]),
        # 
        'MoveFileFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpExistingFileName", "lpNewFileName"]),
        # 
        'RemoveDirectoryFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpPathName"]),
        # 
        'ReplaceFileFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeBottom(label="Void"), offset=0), SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["lpReplacedFileName", "lpReplacementFileName", "lpBackupFileName", "dwReplaceFlags", "lpExclude", "lpReserved"]),
        # 
        'SetFileAttributesFromAppW': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32")], SimTypeInt(signed=True, label="Int32"), arg_names=["lpFileName", "dwFileAttributes"]),
    }

lib.set_prototypes(prototypes)



import logging
import angr

l = logging.getLogger(name=__name__)

######################################
# __initterm
######################################

class _initterm(angr.SimProcedure):
    local_vars = ('callbacks',)
    callbacks = []

    #pylint:disable=arguments-differ
    def run(self, fp_a, fp_z):
        if self.state.solver.symbolic(fp_a) or self.state.solver.symbolic(fp_z):
            l.warning("Symbolic argument to _initterm{_e} is not supported... returning")
            self.ret(0) # might as well try to keep going

        self.callbacks = self.get_callbacks(fp_a, fp_z)
        self.do_callbacks(fp_a, fp_z)

    def get_callbacks(self, fp_a, fp_z):
        callbacks = []
        table_size = fp_z - fp_a + self.state.arch.bytes
        for addr in reversed(self.state.memory.load(fp_a, table_size, endness=self.state.arch.memory_endness).chop(self.state.arch.bits)):
            addr = self.state.solver.eval(addr)
            if addr != 0:
                callbacks.append(addr)
        return callbacks

    def do_callbacks(self, fp_a, fp_z): # pylint:disable=unused-argument
        if len(self.callbacks) == 0:
            self.ret(0)  # probably best to assume each callback returned 0
        else:
            callback_addr = self.callbacks.pop(0)
            l.debug("Calling %#x", callback_addr)
            self.call(callback_addr, [], continue_at='do_callbacks', prototype='void x()')



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("eappcfg.dll")
prototypes = \
    {
        # 
        'EapHostPeerGetMethods': SimTypeFunction([SimTypePointer(SimStruct({"dwNumberOfMethods": SimTypeInt(signed=False, label="UInt32"), "pEapMethods": SimTypePointer(SimStruct({"eaptype": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "pwszAuthorName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszFriendlyName": SimTypePointer(SimTypeChar(label="Char"), offset=0), "eapProperties": SimTypeInt(signed=False, label="UInt32"), "pInnerMethodInfo": SimTypePointer(SimTypeBottom(label="EAP_METHOD_INFO"), offset=0)}, name="EAP_METHOD_INFO", pack=False, align=None), offset=0)}, name="EAP_METHOD_INFO_ARRAY", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["pEapMethodInfoArray", "ppEapError"]),
        # 
        'EapHostPeerGetMethodProperties': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimStruct({"dwNumberOfProperties": SimTypeInt(signed=False, label="UInt32"), "pMethodProperty": SimTypePointer(SimStruct({"eapMethodPropertyType": SimTypeInt(signed=False, label="EAP_METHOD_PROPERTY_TYPE"), "eapMethodPropertyValueType": SimTypeInt(signed=False, label="EAP_METHOD_PROPERTY_VALUE_TYPE"), "eapMethodPropertyValue": SimUnion({"empvBool": SimStruct({"length": SimTypeInt(signed=False, label="UInt32"), "value": SimTypeInt(signed=True, label="Int32")}, name="EAP_METHOD_PROPERTY_VALUE_BOOL", pack=False, align=None), "empvDword": SimStruct({"length": SimTypeInt(signed=False, label="UInt32"), "value": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_PROPERTY_VALUE_DWORD", pack=False, align=None), "empvString": SimStruct({"length": SimTypeInt(signed=False, label="UInt32"), "value": SimTypePointer(SimTypeChar(label="Byte"), offset=0)}, name="EAP_METHOD_PROPERTY_VALUE_STRING", pack=False, align=None)}, name="<anon>", label="None")}, name="EAP_METHOD_PROPERTY", pack=False, align=None), offset=0)}, name="EAP_METHOD_PROPERTY_ARRAY", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwVersion", "dwFlags", "eapMethodType", "hUserImpersonationToken", "dwEapConnDataSize", "pbEapConnData", "dwUserDataSize", "pbUserData", "pMethodPropertyArray", "ppEapError"]),
        # 
        'EapHostPeerInvokeConfigUI': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hwndParent", "dwFlags", "eapMethodType", "dwSizeOfConfigIn", "pConfigIn", "pdwSizeOfConfigOut", "ppConfigOut", "ppEapError"]),
        # 
        'EapHostPeerQueryCredentialInputFields': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwNumberOfFields": SimTypeInt(signed=False, label="UInt32"), "pFields": SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "Type": SimTypeInt(signed=False, label="EAP_CONFIG_INPUT_FIELD_TYPE"), "dwFlagProps": SimTypeInt(signed=False, label="UInt32"), "pwszLabel": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszData": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwMinDataLength": SimTypeInt(signed=False, label="UInt32"), "dwMaxDataLength": SimTypeInt(signed=False, label="UInt32")}, name="EAP_CONFIG_INPUT_FIELD_DATA", pack=False, align=None), offset=0)}, name="EAP_CONFIG_INPUT_FIELD_ARRAY", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hUserImpersonationToken", "eapMethodType", "dwFlags", "dwEapConnDataSize", "pbEapConnData", "pEapConfigInputFieldArray", "ppEapError"]),
        # 
        'EapHostPeerQueryUserBlobFromCredentialInputFields': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwNumberOfFields": SimTypeInt(signed=False, label="UInt32"), "pFields": SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "Type": SimTypeInt(signed=False, label="EAP_CONFIG_INPUT_FIELD_TYPE"), "dwFlagProps": SimTypeInt(signed=False, label="UInt32"), "pwszLabel": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszData": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwMinDataLength": SimTypeInt(signed=False, label="UInt32"), "dwMaxDataLength": SimTypeInt(signed=False, label="UInt32")}, name="EAP_CONFIG_INPUT_FIELD_DATA", pack=False, align=None), offset=0)}, name="EAP_CONFIG_INPUT_FIELD_ARRAY", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), label="LPArray", offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hUserImpersonationToken", "eapMethodType", "dwFlags", "dwEapConnDataSize", "pbEapConnData", "pEapConfigInputFieldArray", "pdwUserBlobSize", "ppbUserBlob", "ppEapError"]),
        # 
        'EapHostPeerInvokeIdentityUI': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwVersion", "eapMethodType", "dwFlags", "hwndParent", "dwSizeofConnectionData", "pConnectionData", "dwSizeofUserData", "pUserData", "pdwSizeOfUserDataOut", "ppUserDataOut", "ppwszIdentity", "ppEapError", "ppvReserved"]),
        # 
        'EapHostPeerInvokeInteractiveUI': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["hwndParent", "dwSizeofUIContextData", "pUIContextData", "pdwSizeOfDataFromInteractiveUI", "ppDataFromInteractiveUI", "ppEapError"]),
        # 
        'EapHostPeerQueryInteractiveUIInputFields': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwSize": SimTypeInt(signed=False, label="UInt32"), "dwDataType": SimTypeInt(signed=False, label="EAP_INTERACTIVE_UI_DATA_TYPE"), "cbUiData": SimTypeInt(signed=False, label="UInt32"), "pbUiData": SimUnion({"credData": SimTypePointer(SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwNumberOfFields": SimTypeInt(signed=False, label="UInt32"), "pFields": SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "Type": SimTypeInt(signed=False, label="EAP_CONFIG_INPUT_FIELD_TYPE"), "dwFlagProps": SimTypeInt(signed=False, label="UInt32"), "pwszLabel": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszData": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwMinDataLength": SimTypeInt(signed=False, label="UInt32"), "dwMaxDataLength": SimTypeInt(signed=False, label="UInt32")}, name="EAP_CONFIG_INPUT_FIELD_DATA", pack=False, align=None), offset=0)}, name="EAP_CONFIG_INPUT_FIELD_ARRAY", pack=False, align=None), offset=0), "credExpiryData": SimTypePointer(SimStruct({"curCreds": SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwNumberOfFields": SimTypeInt(signed=False, label="UInt32"), "pFields": SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "Type": SimTypeInt(signed=False, label="EAP_CONFIG_INPUT_FIELD_TYPE"), "dwFlagProps": SimTypeInt(signed=False, label="UInt32"), "pwszLabel": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszData": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwMinDataLength": SimTypeInt(signed=False, label="UInt32"), "dwMaxDataLength": SimTypeInt(signed=False, label="UInt32")}, name="EAP_CONFIG_INPUT_FIELD_DATA", pack=False, align=None), offset=0)}, name="EAP_CONFIG_INPUT_FIELD_ARRAY", pack=False, align=None), "newCreds": SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwNumberOfFields": SimTypeInt(signed=False, label="UInt32"), "pFields": SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "Type": SimTypeInt(signed=False, label="EAP_CONFIG_INPUT_FIELD_TYPE"), "dwFlagProps": SimTypeInt(signed=False, label="UInt32"), "pwszLabel": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszData": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwMinDataLength": SimTypeInt(signed=False, label="UInt32"), "dwMaxDataLength": SimTypeInt(signed=False, label="UInt32")}, name="EAP_CONFIG_INPUT_FIELD_DATA", pack=False, align=None), offset=0)}, name="EAP_CONFIG_INPUT_FIELD_ARRAY", pack=False, align=None)}, name="EAP_CRED_EXPIRY_REQ", pack=False, align=None), offset=0), "credLogonData": SimTypePointer(SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwNumberOfFields": SimTypeInt(signed=False, label="UInt32"), "pFields": SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "Type": SimTypeInt(signed=False, label="EAP_CONFIG_INPUT_FIELD_TYPE"), "dwFlagProps": SimTypeInt(signed=False, label="UInt32"), "pwszLabel": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszData": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwMinDataLength": SimTypeInt(signed=False, label="UInt32"), "dwMaxDataLength": SimTypeInt(signed=False, label="UInt32")}, name="EAP_CONFIG_INPUT_FIELD_DATA", pack=False, align=None), offset=0)}, name="EAP_CONFIG_INPUT_FIELD_ARRAY", pack=False, align=None), offset=0)}, name="<anon>", label="None")}, name="EAP_INTERACTIVE_UI_DATA", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwVersion", "dwFlags", "dwSizeofUIContextData", "pUIContextData", "pEapInteractiveUIData", "ppEapError", "ppvReserved"]),
        # 
        'EapHostPeerQueryUIBlobFromInteractiveUIInputFields': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwSize": SimTypeInt(signed=False, label="UInt32"), "dwDataType": SimTypeInt(signed=False, label="EAP_INTERACTIVE_UI_DATA_TYPE"), "cbUiData": SimTypeInt(signed=False, label="UInt32"), "pbUiData": SimUnion({"credData": SimTypePointer(SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwNumberOfFields": SimTypeInt(signed=False, label="UInt32"), "pFields": SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "Type": SimTypeInt(signed=False, label="EAP_CONFIG_INPUT_FIELD_TYPE"), "dwFlagProps": SimTypeInt(signed=False, label="UInt32"), "pwszLabel": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszData": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwMinDataLength": SimTypeInt(signed=False, label="UInt32"), "dwMaxDataLength": SimTypeInt(signed=False, label="UInt32")}, name="EAP_CONFIG_INPUT_FIELD_DATA", pack=False, align=None), offset=0)}, name="EAP_CONFIG_INPUT_FIELD_ARRAY", pack=False, align=None), offset=0), "credExpiryData": SimTypePointer(SimStruct({"curCreds": SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwNumberOfFields": SimTypeInt(signed=False, label="UInt32"), "pFields": SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "Type": SimTypeInt(signed=False, label="EAP_CONFIG_INPUT_FIELD_TYPE"), "dwFlagProps": SimTypeInt(signed=False, label="UInt32"), "pwszLabel": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszData": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwMinDataLength": SimTypeInt(signed=False, label="UInt32"), "dwMaxDataLength": SimTypeInt(signed=False, label="UInt32")}, name="EAP_CONFIG_INPUT_FIELD_DATA", pack=False, align=None), offset=0)}, name="EAP_CONFIG_INPUT_FIELD_ARRAY", pack=False, align=None), "newCreds": SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwNumberOfFields": SimTypeInt(signed=False, label="UInt32"), "pFields": SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "Type": SimTypeInt(signed=False, label="EAP_CONFIG_INPUT_FIELD_TYPE"), "dwFlagProps": SimTypeInt(signed=False, label="UInt32"), "pwszLabel": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszData": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwMinDataLength": SimTypeInt(signed=False, label="UInt32"), "dwMaxDataLength": SimTypeInt(signed=False, label="UInt32")}, name="EAP_CONFIG_INPUT_FIELD_DATA", pack=False, align=None), offset=0)}, name="EAP_CONFIG_INPUT_FIELD_ARRAY", pack=False, align=None)}, name="EAP_CRED_EXPIRY_REQ", pack=False, align=None), offset=0), "credLogonData": SimTypePointer(SimStruct({"dwVersion": SimTypeInt(signed=False, label="UInt32"), "dwNumberOfFields": SimTypeInt(signed=False, label="UInt32"), "pFields": SimTypePointer(SimStruct({"dwSize": SimTypeInt(signed=False, label="UInt32"), "Type": SimTypeInt(signed=False, label="EAP_CONFIG_INPUT_FIELD_TYPE"), "dwFlagProps": SimTypeInt(signed=False, label="UInt32"), "pwszLabel": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pwszData": SimTypePointer(SimTypeChar(label="Char"), offset=0), "dwMinDataLength": SimTypeInt(signed=False, label="UInt32"), "dwMaxDataLength": SimTypeInt(signed=False, label="UInt32")}, name="EAP_CONFIG_INPUT_FIELD_DATA", pack=False, align=None), offset=0)}, name="EAP_CONFIG_INPUT_FIELD_ARRAY", pack=False, align=None), offset=0)}, name="<anon>", label="None")}, name="EAP_INTERACTIVE_UI_DATA", pack=False, align=None), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeBottom(label="Void"), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwVersion", "dwFlags", "dwSizeofUIContextData", "pUIContextData", "pEapInteractiveUIData", "pdwSizeOfDataFromInteractiveUI", "ppDataFromInteractiveUI", "ppEapError", "ppvReserved"]),
        # 
        'EapHostPeerConfigXml2Blob': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeBottom(label="IXMLDOMNode"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0), SimTypePointer(SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwFlags", "pConfigDoc", "pdwSizeOfConfigOut", "ppConfigOut", "pEapMethodType", "ppEapError"]),
        # 
        'EapHostPeerCredentialsXml2Blob': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimTypeBottom(label="IXMLDOMNode"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Byte"), offset=0), offset=0), SimTypePointer(SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwFlags", "pCredentialsDoc", "dwSizeOfConfigIn", "pConfigIn", "pdwSizeOfCredentialsOut", "ppCredentialsOut", "pEapMethodType", "ppEapError"]),
        # 
        'EapHostPeerConfigBlob2Xml': SimTypeFunction([SimTypeInt(signed=False, label="UInt32"), SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypeChar(label="Byte"), label="LPArray", offset=0), SimTypePointer(SimTypeBottom(label="IXMLDOMDocument2"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0), offset=0)], SimTypeInt(signed=False, label="UInt32"), arg_names=["dwFlags", "eapMethodType", "dwSizeOfConfigIn", "pConfigIn", "ppConfigDoc", "ppEapError"]),
        # 
        'EapHostPeerFreeMemory': SimTypeFunction([SimTypePointer(SimTypeChar(label="Byte"), offset=0)], SimTypeBottom(label="Void"), arg_names=["pData"]),
        # 
        'EapHostPeerFreeErrorMemory': SimTypeFunction([SimTypePointer(SimStruct({"dwWinError": SimTypeInt(signed=False, label="UInt32"), "type": SimStruct({"eapType": SimStruct({"type": SimTypeChar(label="Byte"), "dwVendorId": SimTypeInt(signed=False, label="UInt32"), "dwVendorType": SimTypeInt(signed=False, label="UInt32")}, name="EAP_TYPE", pack=False, align=None), "dwAuthorId": SimTypeInt(signed=False, label="UInt32")}, name="EAP_METHOD_TYPE", pack=False, align=None), "dwReasonCode": SimTypeInt(signed=False, label="UInt32"), "rootCauseGuid": SimTypeBottom(label="Guid"), "repairGuid": SimTypeBottom(label="Guid"), "helpLinkGuid": SimTypeBottom(label="Guid"), "pRootCauseString": SimTypePointer(SimTypeChar(label="Char"), offset=0), "pRepairString": SimTypePointer(SimTypeChar(label="Char"), offset=0)}, name="EAP_ERROR", pack=False, align=None), offset=0)], SimTypeBottom(label="Void"), arg_names=["pEapError"]),
    }

lib.set_prototypes(prototypes)



import angr
import claripy
import logging

l = logging.getLogger(name=__name__)

class LoadLibraryA(angr.SimProcedure):
    def run(self, lib_ptr):
        lib = self.state.mem[lib_ptr].string.concrete.decode('utf-8')
        return self.load(lib)

    def load(self, lib):
        if '.' not in lib:
            lib += '.dll'
        loaded = self.project.loader.dynamic_load(lib)
        if loaded is None:
            l.debug("LoadLibrary: Could not load %s", lib)
            return 0

        # Add simprocedures
        for obj in loaded:
            self.register(obj)

        l.debug("LoadLibrary: Loaded %s", lib)
        return self.project.loader.find_object(lib).mapped_base

    def register(self, obj): # can be overridden for instrumentation
        self.project._register_object(obj, obj.arch)

class LoadLibraryExW(LoadLibraryA):
    def run(self, lib_ptr, flag1, flag2):
        lib = self.state.mem[lib_ptr].wstring.concrete
        return self.load(lib)

# if you subclass LoadLibraryA to provide register, you can implement LoadLibraryExW by making an empty class that just
# subclasses your special procedure and LoadLibraryExW


class GetProcAddress(angr.SimProcedure):
    def run(self, lib_handle, name_addr):
        if lib_handle.symbolic:
            raise angr.errors.SimValueError("GetProcAddress called with symbolic library handle %s" % lib_handle)
        lib_handle = self.state.solver.eval(lib_handle)

        if lib_handle == 0:
            obj = self.project.loader.main_object
        else:
            for obj in self.project.loader.all_pe_objects:
                if obj.mapped_base == lib_handle:
                    break
            else:
                l.warning("GetProcAddress: invalid library handle %s", lib_handle)
                return 0

        if claripy.is_true(name_addr < 0x10000):
            # this matches the bogus name specified in the loader...
            ordinal = self.state.solver.eval(name_addr)
            name = 'ordinal.%d.%s' % (ordinal, obj.provides)
        else:
            name = self.state.mem[name_addr].string.concrete.decode('utf-8')

        full_name = '%s.%s' % (obj.provides, name)
        self.procs.add(full_name)

        sym = obj.get_symbol(name)
        if sym is None and name.endswith('@'):
            # There seems to be some mangling parsing being done in the linker?
            # I don't know what I'm doing
            for suffix in ['Z', 'XZ']:
                sym = obj.get_symbol(name + suffix)
                if sym is not None:
                    name = name + suffix
                    break

        if sym is None:
            l.info("GetProcAddress: object %s does not contain %s", obj.provides, name)
            return 0

        sym = sym.resolve_forwarder()
        if sym is None:
            l.warning("GetProcAddress: forwarding failed for %s from %s", name, obj.provides)
            return 0

        name = sym.name # fix ordinal names
        full_name = '%s.%s' % (obj.provides, name)
        self.procs.add(full_name)

        l.debug("GetProcAddress: Imported %s (%#x) from %s", name, sym.rebased_addr, obj.provides)
        return sym.rebased_addr

    KEY = 'dynamically_loaded_procedures'
    @property
    def procs(self):
        try:
            return self.state.globals[self.KEY]
        except KeyError:
            x = set()
            self.state.globals[self.KEY] = x
            return x



from typing import Dict, Tuple, Union, Set

import claripy

from .undefined import Undefined, UNDEFINED


class Environment:
    """
    Represent the environment in which a program runs.
    It's a mapping of variable names, to `claripy.ast.Base` that should contain possible addresses, or <UNDEFINED>, at
    which their respective values are stored.

    **Note**: The <Environment> object does not store the values associated with variables themselves.
    """
    def __init__(self, environment: Dict[Union[str,Undefined],Set[claripy.ast.Base]]=None):
        self._environment: Dict[Union[str,Undefined],Set[claripy.ast.Base]] = environment or {}

    def get(self, names: Set[str]) -> Tuple[Set[claripy.ast.Base],bool]:
        """
        :param names: Potential values for the name of the environment variable to get the pointers of.
        :return:
            The potential addresses of the values the environment variable can take;
            And a boolean value telling whether all the names were known of the internal representation (i.e. will be
            False if one of the queried variable was not found).
        """
        has_unknown = not all(map(lambda name: name in self._environment.keys(), names))

        def _get(name):
            if not isinstance(name, (str, Undefined)):
                raise TypeError("get(): Expected str, or Undefined, got %s" % type(name).__name__)
            return self._environment.get(name, {UNDEFINED})

        pointers = set()
        for values in map(_get, names):
            pointers |= values

        return pointers, has_unknown

    def set(self, name: Union[str,Undefined], pointers: Set[claripy.ast.Base]):
        """
        :param name: Name of the environment variable to which we will associate the pointers.
        :param pointers: New addresses where the new values of the environment variable are located.
        """
        if not isinstance(name, (str, Undefined)):
            raise TypeError("set(): Expected str, or Undefined, got %s" % type(name).__name__)
        self._environment[name] = pointers

    def __str__(self):
        return "Environment: %s" % self._environment

    def __repr__(self):
        return "Environment: %s" % self._environment

    def __eq__(self, other: object) -> bool:
        assert isinstance(other, Environment), "Cannot compare Environment with %s" % type(other).__name__
        return self._environment == other._environment

    def merge(self, *others: 'Environment') -> Tuple['Environment',bool]:

        new_env = self._environment

        for other in others:
            if not isinstance(other, Environment):
                raise TypeError("Cannot merge Environment with %s" % type(other).__name__)

            keys = set(new_env.keys())
            keys |= other._environment.keys()

            def _dataset_from_key(key, environment1, environment2):
                v = environment1.get(key, None)
                w = environment2.get(key, None)
                # Because the key is coming from one of them, they cannot be both `None`.
                if v is None: return w
                if w is None: return v
                return v | w

            new_env = dict(map(
                lambda k: (k, _dataset_from_key(k, new_env, other._environment)),
                keys
            ))

        merge_occurred = new_env != self._environment
        return Environment(environment=new_env), merge_occurred



import sys

def detect_test_env():
    i = 0
    while True:
        i += 1
        try:
            frame_module = sys._getframe(i).f_globals.get('__name__')
        except ValueError:
            return False

        if frame_module == '__main__' or frame_module == '__console__':
            return False
        elif frame_module is not None and (frame_module.startswith('nose.') or frame_module.startswith('nose2.')):
            return True

is_testing = detect_test_env()



from typing import Optional, Tuple, Any

import networkx

import ailment

from .structurer_nodes import MultiNode, BaseNode, CodeNode, SequenceNode, ConditionNode, SwitchCaseNode, \
    CascadingConditionNode


def remove_last_statement(node):
    stmt = None

    if type(node) is CodeNode:
        stmt = remove_last_statement(node.node)
    elif type(node) is ailment.Block:
        stmt = node.statements[-1]
        node.statements = node.statements[:-1]
    elif type(node) is MultiNode:
        if node.nodes:
            stmt = remove_last_statement(node.nodes[-1])
            if BaseNode.test_empty_node(node.nodes[-1]):
                node.nodes = node.nodes[:-1]
    elif type(node) is SequenceNode:
        if node.nodes:
            stmt = remove_last_statement(node.nodes[-1])
            if BaseNode.test_empty_node(node.nodes[-1]):
                node.nodes = node.nodes[:-1]
    else:
        raise NotImplementedError()

    return stmt


def append_statement(node, stmt):

    if type(node) is CodeNode:
        append_statement(node.node, stmt)
        return
    if type(node) is ailment.Block:
        node.statements.append(stmt)
        return
    if type(node) is MultiNode:
        if node.nodes:
            append_statement(node.nodes[-1], stmt)
        else:
            raise NotImplementedError()
        return
    if type(node) is SequenceNode:
        if node.nodes:
            append_statement(node.nodes[-1], stmt)
        else:
            raise NotImplementedError()
        return

    raise NotImplementedError()


def replace_last_statement(node, old_stmt, new_stmt):

    if type(node) is CodeNode:
        replace_last_statement(node.node, old_stmt, new_stmt)
        return
    if type(node) is ailment.Block:
        if node.statements[-1] is old_stmt:
            node.statements[-1] = new_stmt
        return
    if type(node) is MultiNode:
        if node.nodes:
            replace_last_statement(node.nodes[-1], old_stmt, new_stmt)
        return
    if type(node) is SequenceNode:
        if node.nodes:
            replace_last_statement(node.nodes[-1], old_stmt, new_stmt)
        return
    if type(node) is ConditionNode:
        if node.true_node is not None:
            replace_last_statement(node.true_node, old_stmt, new_stmt)
        if node.false_node is not None:
            replace_last_statement(node.false_node, old_stmt, new_stmt)
        return

    raise NotImplementedError()


def extract_jump_targets(stmt):
    """
    Extract concrete goto targets from a Jump or a ConditionalJump statement.

    :param stmt:    The statement to analyze.
    :return:        A list of known concrete jump targets.
    :rtype:         list
    """

    targets = [ ]

    if isinstance(stmt, ailment.Stmt.Jump):
        if isinstance(stmt.target, ailment.Expr.Const):
            targets.append(stmt.target.value)
    elif isinstance(stmt, ailment.Stmt.ConditionalJump):
        if isinstance(stmt.true_target, ailment.Expr.Const):
            targets.append(stmt.true_target.value)
        if isinstance(stmt.false_target, ailment.Expr.Const):
            targets.append(stmt.false_target.value)

    return targets


def switch_extract_cmp_bounds(last_stmt: ailment.Stmt.ConditionalJump) -> Optional[Tuple[Any,int,int]]:
    """
    Check the last statement of the switch-case header node, and extract lower+upper bounds for the comparison.

    :param last_stmt:   The last statement of the switch-case header node.
    :return:            A tuple of (comparison expression, lower bound, upper bound), or None
    """

    if not isinstance(last_stmt, ailment.Stmt.ConditionalJump):
        return None

    # TODO: Add more operations
    if isinstance(last_stmt.condition, ailment.Expr.BinaryOp) and last_stmt.condition.op == 'CmpLE':
        if not isinstance(last_stmt.condition.operands[1], ailment.Expr.Const):
            return None
        cmp_ub = last_stmt.condition.operands[1].value
        cmp_lb = 0
        cmp = last_stmt.condition.operands[0]
        if isinstance(cmp, ailment.Expr.BinaryOp) and \
                cmp.op == 'Sub' and \
                isinstance(cmp.operands[1], ailment.Expr.Const):
            cmp_ub += cmp.operands[1].value
            cmp_lb += cmp.operands[1].value
            cmp = cmp.operands[0]
        return cmp, cmp_lb, cmp_ub

    return None


def get_ast_subexprs(claripy_ast):

    queue = [ claripy_ast ]
    while queue:
        ast = queue.pop(0)
        if ast.op == "And":
            queue += ast.args[1:]
            yield ast.args[0]
        else:
            yield ast


def insert_node(parent, insert_idx, node, node_idx, label=None, insert_location=None):

    if isinstance(parent, SequenceNode):
        parent.nodes.insert(insert_idx, node)
    elif isinstance(parent, CodeNode):
        # Make a new sequence node
        seq = SequenceNode(parent.addr, nodes=[parent.node, node])
        parent.node = seq
    elif isinstance(parent, MultiNode):
        parent.nodes.insert(insert_idx, node)
    elif isinstance(parent, ConditionNode):
        if node_idx == 0:
            # true node
            if not isinstance(parent.true_node, (SequenceNode, MultiNode)):
                if parent.true_node is None:
                    parent.true_node = SequenceNode(parent.addr, nodes=[])
                else:
                    parent.true_node = SequenceNode(parent.true_node.addr, nodes=[parent.true_node])
            insert_node(parent.true_node, insert_idx - node_idx, node, 0)
        else:
            # false node
            if not isinstance(parent.false_node, (SequenceNode, MultiNode)):
                if parent.false_node is None:
                    parent.false_node = SequenceNode(parent.addr, nodes=[])
                else:
                    parent.false_node = SequenceNode(parent.false_node.addr, nodes=[parent.false_node])
            insert_node(parent.false_node, insert_idx - node_idx, node, 0)
    elif isinstance(parent, CascadingConditionNode):
        cond, child_node = parent.condition_and_nodes[node_idx]
        if not isinstance(child_node, SequenceNode):
            child_node = SequenceNode(child_node.addr, nodes=[child_node])
            parent.condition_and_nodes[node_idx] = (cond, child_node)
        insert_node(child_node, insert_idx - node_idx, node, 0)
    elif isinstance(parent, SwitchCaseNode):
        # note that this case will be hit only when the parent node is not a container, such as SequenceNode or
        # MultiNode. we always need to create a new SequenceNode and replace the original node in place.
        if label == 'switch_expr':
            raise TypeError("You cannot insert a node after an expression.")
        if label == 'case':
            # node_idx is the case number
            if insert_location == 'after':
                new_nodes = [ parent.cases[node_idx], node ]
            elif insert_location == 'before':
                new_nodes = [ node, parent.cases[node_idx] ]
            else:
                raise TypeError("Unsupported 'insert_location' value %r." % insert_location)
            seq = SequenceNode(new_nodes[0].addr, nodes=new_nodes)
            parent.cases[node_idx] = seq
        elif label == 'default':
            if insert_location == 'after':
                new_nodes = [ parent.default_node, node ]
            elif insert_location == 'before':
                new_nodes = [ node, parent.default_node ]
            else:
                raise TypeError("Unsupported 'insert_location' value %r." % insert_location)
            seq = SequenceNode(new_nodes[0].addr, nodes=new_nodes)
            parent.default_node = seq
    else:
        raise NotImplementedError()


def _merge_ail_nodes(graph, node_a: ailment.Block, node_b: ailment.Block) -> ailment.Block:
    in_edges = list(graph.in_edges(node_a, data=True))
    out_edges = list(graph.out_edges(node_b, data=True))

    new_node = node_a.copy() if node_a.addr <= node_b.addr else node_b.copy()
    old_node = node_b if new_node == node_a else node_a
    new_node.statements += old_node.statements
    new_node.original_size += old_node.original_size

    graph.remove_node(node_a)
    graph.remove_node(node_b)

    if new_node is not None:
        graph.add_node(new_node)

        for src, _, data in in_edges:
            if src is node_b:
                src = new_node
            graph.add_edge(src, new_node, **data)

        for _, dst, data in out_edges:
            if dst is node_a:
                dst = new_node
            graph.add_edge(new_node, dst, **data)

    return new_node


def to_ail_supergraph(transition_graph: networkx.DiGraph) -> networkx.DiGraph:
    """
    Takes an AIL graph and converts it into a AIL graph that treats calls and redundant jumps
    as parts of a bigger block instead of transitions. Calls to returning functions do not terminate basic blocks.

    Based on region_identifier super_graph

    :return: A converted super transition graph
    """
    # make a copy of the graph
    transition_graph = networkx.DiGraph(transition_graph)

    while True:
        for src, dst, data in transition_graph.edges(data=True):
            type_ = data.get('type', None)

            if len(list(transition_graph.successors(src))) == 1 and len(list(transition_graph.predecessors(dst))) == 1:
                # calls in the middle of blocks OR boring jumps
                if (type_ == 'fake_return') or (src.addr + src.original_size == dst.addr):
                    _merge_ail_nodes(transition_graph, src, dst)
                    break

            # calls to functions with no return
            elif type_ == 'call':
                transition_graph.remove_node(dst)
                break
        else:
            break

    return transition_graph



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: protos/cfg.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import primitives_pb2 as protos_dot_primitives__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10protos/cfg.proto\x12\x0b\x61ngr.protos\x1a\x17protos/primitives.proto\"H\n\x07\x43\x46GNode\x12\n\n\x02\x65\x61\x18\x01 \x01(\x04\x12\x0c\n\x04size\x18\x02 \x01(\r\x12\x10\n\x08\x62lock_id\x18\x03 \x03(\x03\x12\x11\n\treturning\x18\x04 \x01(\x08\"\x89\x01\n\x03\x43\x46G\x12\r\n\x05ident\x18\x01 \x01(\t\x12#\n\x05nodes\x18\x02 \x03(\x0b\x32\x14.angr.protos.CFGNode\x12 \n\x05\x65\x64ges\x18\x03 \x03(\x0b\x32\x11.angr.protos.Edge\x12,\n\x0bmemory_data\x18\x04 \x03(\x0b\x32\x17.angr.protos.MemoryData\"\xae\x02\n\nMemoryData\x12\n\n\x02\x65\x61\x18\x01 \x01(\x04\x12\x0c\n\x04size\x18\x02 \x01(\r\x12\x34\n\x04type\x18\x03 \x01(\x0e\x32&.angr.protos.MemoryData.MemoryDataType\"\xcf\x01\n\x0eMemoryDataType\x12\x13\n\x0fUnknownDataType\x10\x00\x12\x0f\n\x0bUnspecified\x10\x01\x12\x0b\n\x07Integer\x10\x02\x12\x10\n\x0cPointerArray\x10\x03\x12\n\n\x06String\x10\x04\x12\x11\n\rUnicodeString\x10\x05\x12\x13\n\x0fSegmentBoundary\x10\x06\x12\x11\n\rCodeReference\x10\x07\x12\x0f\n\x0bGOTPLTEntry\x10\x08\x12\r\n\tELFHeader\x10\t\x12\x11\n\rFloatingPoint\x10\nb\x06proto3')



_CFGNODE = DESCRIPTOR.message_types_by_name['CFGNode']
_CFG = DESCRIPTOR.message_types_by_name['CFG']
_MEMORYDATA = DESCRIPTOR.message_types_by_name['MemoryData']
_MEMORYDATA_MEMORYDATATYPE = _MEMORYDATA.enum_types_by_name['MemoryDataType']
CFGNode = _reflection.GeneratedProtocolMessageType('CFGNode', (_message.Message,), {
  'DESCRIPTOR' : _CFGNODE,
  '__module__' : 'protos.cfg_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.CFGNode)
  })
_sym_db.RegisterMessage(CFGNode)

CFG = _reflection.GeneratedProtocolMessageType('CFG', (_message.Message,), {
  'DESCRIPTOR' : _CFG,
  '__module__' : 'protos.cfg_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.CFG)
  })
_sym_db.RegisterMessage(CFG)

MemoryData = _reflection.GeneratedProtocolMessageType('MemoryData', (_message.Message,), {
  'DESCRIPTOR' : _MEMORYDATA,
  '__module__' : 'protos.cfg_pb2'
  # @@protoc_insertion_point(class_scope:angr.protos.MemoryData)
  })
_sym_db.RegisterMessage(MemoryData)

if _descriptor._USE_C_DESCRIPTORS == False:

  DESCRIPTOR._options = None
  _CFGNODE._serialized_start=58
  _CFGNODE._serialized_end=130
  _CFG._serialized_start=133
  _CFG._serialized_end=270
  _MEMORYDATA._serialized_start=273
  _MEMORYDATA._serialized_end=575
  _MEMORYDATA_MEMORYDATATYPE._serialized_start=368
  _MEMORYDATA_MEMORYDATATYPE._serialized_end=575
# @@protoc_insertion_point(module_scope)



# pylint:disable=arguments-differ,arguments-renamed,isinstance-second-argument-not-valid-type
from typing import Optional, Union, TYPE_CHECKING
import logging

import claripy
from ailment import Stmt, Expr

from ...utils.constants import is_alignment_mask
from ...engines.light import SimEngineLightAILMixin
from ...sim_variable import SimStackVariable, SimMemoryVariable
from .engine_base import SimEnginePropagatorBase
from .prop_value import PropValue, Detail

if TYPE_CHECKING:
    from .propagator import PropagatorAILState
    from angr.code_location import CodeLocation

l = logging.getLogger(name=__name__)


class SimEnginePropagatorAIL(
    SimEngineLightAILMixin,
    SimEnginePropagatorBase,
):
    """
    The AIl engine for Propagator.
    """

    state: 'PropagatorAILState'

    def _is_top(self, expr: Union[claripy.ast.Base,Expr.StackBaseOffset]) -> bool:
        if isinstance(expr, Expr.StackBaseOffset):
            return False
        return super()._is_top(expr)

    def extract_offset_to_sp(self, expr: Union[claripy.ast.Base,Expr.StackBaseOffset]) -> Optional[int]:
        if isinstance(expr, Expr.StackBaseOffset):
            return expr.offset
        elif isinstance(expr, Expr.Expression):
            # not supported
            return None
        return super().extract_offset_to_sp(expr)

    #
    # AIL statement handlers
    #

    def _ail_handle_Assignment(self, stmt):
        """

        :param Stmt.Assignment stmt:
        :return:
        """

        src = self._expr(stmt.src)
        dst = stmt.dst

        if type(dst) is Expr.Tmp:
            self.state.store_temp(dst.tmp_idx, src)
            self.state.temp_expressions[dst.tmp_idx] = stmt.src

        elif type(dst) is Expr.Register:
            if src.needs_details:
                # provide details
                src = src.with_details(dst.size, dst, self._codeloc())

            self.state.store_register(dst, src)
            if isinstance(stmt.src, (Expr.Register, Stmt.Call)):
                # set equivalence
                self.state.add_equivalence(self._codeloc(), dst, stmt.src)
            if isinstance(stmt.src, (Expr.Convert)) and isinstance(stmt.src.operand, Stmt.Call):
                # set equivalence
                self.state.add_equivalence(self._codeloc(), dst, stmt.src)

            if src.one_expr is not None:
                self.state.register_expressions[(dst.reg_offset, dst.size)] = dst, src.one_expr, self._codeloc()
            else:
                self.state.register_expressions[(dst.reg_offset, dst.size)] = dst, stmt.src, self._codeloc()
        else:
            l.warning('Unsupported type of Assignment dst %s.', type(dst).__name__)

    def _ail_handle_Store(self, stmt: Stmt.Store):

        self.state: 'PropagatorAILState'

        addr = self._expr(stmt.addr)
        data = self._expr(stmt.data)

        # is it accessing the stack?
        sp_offset = self.extract_offset_to_sp(addr.one_expr) if addr.one_expr is not None else None
        if sp_offset is not None:
            if isinstance(data.one_expr, Expr.StackBaseOffset):
                # convert it to a BV
                expr = data.one_expr
                data_v = self.sp_offset(data.one_expr.offset)
                size = data_v.size() // self.arch.byte_width
                to_store = PropValue.from_value_and_details(data_v, size, expr, self._codeloc())
            elif isinstance(data.value, claripy.ast.BV):
                expr = data.one_expr if data.one_expr is not None else stmt.data
                data_v = data.value
                size = data_v.size() // self.arch.byte_width
                to_store = PropValue.from_value_and_details(data_v, size, expr, self._codeloc())
            else:
                size = stmt.size
                to_store = data.with_details(stmt.size, data.one_expr if data.one_expr is not None else stmt.data,
                                             self._codeloc())

            # Storing data to a stack variable
            self.state.store_stack_variable(sp_offset, to_store, endness=stmt.endness)

            # set equivalence
            var = SimStackVariable(sp_offset, size)
            self.state.add_equivalence(self._codeloc(), var, stmt.data)

        else:
            addr_concrete = addr.one_expr
            if addr_concrete is None:
                # it can be a potential stack store with a variable offset
                self.state.last_stack_store = (self.block.addr, self.stmt_idx, stmt)
            else:
                self.state.global_stores.append((self.block.addr, self.stmt_idx, addr_concrete, stmt))
                if isinstance(addr_concrete, Expr.Const) and isinstance(stmt.size, int):
                    # set equivalence
                    var = SimMemoryVariable(addr_concrete.value, stmt.size)
                    self.state.add_equivalence(self._codeloc(), var, stmt.data)

    def _ail_handle_Jump(self, stmt):
        target = self._expr(stmt.target)
        if target is None or target.one_expr == stmt.target:
            return

        target_oneexpr = target.one_expr
        if target_oneexpr is not None and isinstance(target_oneexpr, Expr.Const):
            new_jump_stmt = Stmt.Jump(stmt.idx, target.one_expr, **stmt.tags)
            self.state.add_replacement(self._codeloc(),
                                       stmt,
                                       new_jump_stmt,
                                       )

    def _ail_handle_Call(self, expr_stmt: Stmt.Call):
        if isinstance(expr_stmt.target, Expr.Expression):
            _ = self._expr(expr_stmt.target)

        if expr_stmt.args:
            for arg in expr_stmt.args:
                _ = self._expr(arg)

        if expr_stmt.ret_expr is not None:
            if isinstance(expr_stmt.ret_expr, Expr.Register):
                # it has a return expression. awesome - treat it as an assignment

                # assume the return value always uses a full-width register
                # FIXME: Expose it as a configuration option
                return_value_use_full_width_reg = True
                if return_value_use_full_width_reg:
                    v = PropValue.from_value_and_details(
                        self.state.top(self.arch.bits), self.arch.bytes, expr_stmt.ret_expr, self._codeloc()
                    )
                    self.state.store_register(
                        Expr.Register(None, expr_stmt.ret_expr.variable, expr_stmt.ret_expr.reg_offset, self.arch.bits,
                                      reg_name=self.arch.translate_register_name(expr_stmt.ret_expr.reg_offset,
                                                                                 size=self.arch.bits)),
                        v
                    )
                else:
                    v = PropValue.from_value_and_details(
                        self.state.top(expr_stmt.ret_expr.size * self.arch.byte_width),
                        expr_stmt.ret_expr.size, expr_stmt.ret_expr, self._codeloc()
                    )
                    self.state.store_register(expr_stmt.ret_expr, v)
                # set equivalence
                self.state.add_equivalence(self._codeloc(), expr_stmt.ret_expr, expr_stmt)
            else:
                l.warning("Unsupported ret_expr type %s.", expr_stmt.ret_expr.__class__)

    def _ail_handle_ConditionalJump(self, stmt):
        condition = self._expr(stmt.condition)
        if stmt.true_target is not None:
            true_target = self._expr(stmt.true_target)
        else:
            true_target = None
        if stmt.false_target is not None:
            false_target = self._expr(stmt.false_target)
        else:
            false_target = None

        # parse the condition to set initial values for true/false branches
        if condition is not None and isinstance(true_target.one_expr, Expr.Const):
            cond_expr = condition.one_expr
            if isinstance(cond_expr, Expr.BinaryOp) and cond_expr.op == "CmpEQ":
                if isinstance(cond_expr.operands[1], Expr.Const):
                    # is there a register that's equivalent to the variable?
                    for _, (reg_atom, reg_expr, def_at) in self.state.register_expressions.items():
                        if cond_expr.operands[0] == reg_expr:
                            # found it!
                            key = self.block.addr, true_target.one_expr.value
                            self.state.block_initial_reg_values[key].append((
                                reg_atom,
                                cond_expr.operands[1],
                            ))


    def _ail_handle_Return(self, stmt: Stmt.Return):
        if stmt.ret_exprs:
            for ret_expr in stmt.ret_exprs:
                self._expr(ret_expr)

    #
    # AIL expression handlers
    #

    # this method exists so that I can annotate the return type
    def _expr(self, expr) -> Optional[PropValue]:  # pylint:disable=useless-super-delegation
        return super()._expr(expr)

    def _ail_handle_Tmp(self, expr: Expr.Tmp) -> PropValue:
        tmp = self.state.load_tmp(expr.tmp_idx)

        if tmp is not None:
            # very first step - if we can get rid of this tmp and replace it with another, we should
            if expr.tmp_idx in self.state.temp_expressions:
                tmp_expr = self.state.temp_expressions[expr.tmp_idx]
                for _, (reg_atom, reg_expr, def_at) in self.state.register_expressions.items():
                    if reg_expr.likes(tmp_expr):
                        # make sure the register still holds the same value
                        current_reg_value = self.state.load_register(reg_atom)
                        if current_reg_value is not None:
                            if 0 in current_reg_value.offset_and_details:
                                detail = current_reg_value.offset_and_details[0]
                                if detail.def_at == def_at:
                                    l.debug("Add a replacement: %s with %s", expr, reg_atom)
                                    self.state.add_replacement(self._codeloc(), expr, reg_atom)
                                    top = self.state.top(expr.size * self.arch.byte_width)
                                    return PropValue.from_value_and_details(top, expr.size, expr, self._codeloc())

            # check if this new_expr uses any expression that has been overwritten
            all_subexprs = list(tmp.all_exprs())
            outdated = False
            for detail in tmp.offset_and_details.values():
                if detail.expr is None:
                    continue
                if self.is_using_outdated_def(detail.expr, detail.def_at, avoid=expr):
                    outdated = True
                    break

            if None in all_subexprs or outdated:
                top = self.state.top(expr.size * self.arch.byte_width)
                self.state.add_replacement(self._codeloc(), expr, top)
                return PropValue.from_value_and_details(top, expr.size, expr, self._codeloc())

            if len(all_subexprs) == 1 and 0 in tmp.offset_and_details and tmp.offset_and_details[0].size == expr.size:
                subexpr = all_subexprs[0]
                l.debug("Add a replacement: %s with %s", expr, subexpr)
                self.state.add_replacement(self._codeloc(), expr, subexpr)
            elif tmp.offset_and_details and 0 in tmp.offset_and_details:
                non_zero_subexprs = list(tmp.non_zero_exprs())
                if len(non_zero_subexprs) == 1 and non_zero_subexprs[0] is tmp.offset_and_details[0].expr:
                    # we will use the zero-extended version as the replacement
                    subexpr = non_zero_subexprs[0]
                    subexpr = PropValue.extend_ail_expression(expr.bits - subexpr.bits, subexpr)
                    l.debug("Add a replacement: %s with %s", expr, subexpr)
                    self.state.add_replacement(self._codeloc(), expr, subexpr)
            return tmp

        if not self._propagate_tmps:
            # we should not propagate any tmps. as a result, we return None for reading attempts to a tmp.
            return PropValue(self.state.top(expr.size * self.arch.byte_width))

        return PropValue(self.state.top(expr.size * self.arch.byte_width))

    def _ail_handle_Register(self, expr: Expr.Register) -> Optional[PropValue]:

        self.state: 'PropagatorAILState'

        # Special handling for SP and BP
        if self._stack_pointer_tracker is not None:
            if expr.reg_offset == self.arch.sp_offset:
                sb_offset = self._stack_pointer_tracker.offset_before(self.ins_addr, self.arch.sp_offset)
                if sb_offset is not None:
                    new_expr = Expr.StackBaseOffset(None, self.arch.bits, sb_offset)
                    self.state.add_replacement(self._codeloc(), expr, new_expr)
                    return PropValue.from_value_and_details(
                        self.sp_offset(sb_offset), expr.size, new_expr, self._codeloc()
                    )
            elif expr.reg_offset == self.arch.bp_offset:
                sb_offset = self._stack_pointer_tracker.offset_before(self.ins_addr, self.arch.bp_offset)
                if sb_offset is not None:
                    new_expr = Expr.StackBaseOffset(None, self.arch.bits, sb_offset)
                    self.state.add_replacement(self._codeloc(), expr, new_expr)
                    return PropValue.from_value_and_details(
                        self.sp_offset(sb_offset), expr.size, new_expr, self._codeloc()
                    )

        def _test_concatenation(pv: PropValue):
            if pv.offset_and_details is not None and len(pv.offset_and_details) == 2 and 0 in pv.offset_and_details:
                lo_value = pv.offset_and_details[0]
                hi_offset = next(iter(k for k in pv.offset_and_details if k != 0))
                hi_value = pv.offset_and_details[hi_offset]
                if lo_value.def_at == hi_value.def_at:
                    # it's the same value! we can apply concatenation here
                    if isinstance(hi_value.expr, Expr.Const) and hi_value.expr.value == 0:
                        # it's probably an up-cast
                        mappings = {
                            # (lo_value.size, hi_value.size): (from_bits, to_bits)
                            (1, 1): (8, 16),  # char to short
                            (1, 3): (8, 32),  # char to int
                            (1, 7): (8, 64),  # char to int64
                            (2, 2): (16, 32),  # short to int
                            (2, 6): (16, 64),  # short to int64
                            (4, 4): (32, 64),  # int to int64
                        }
                        key = (lo_value.size, hi_value.size)
                        if key in mappings:
                            from_bits, to_bits = mappings[key]
                            result_expr = Expr.Convert(None, from_bits, to_bits, False, lo_value.expr)
                            return True, result_expr
                    result_expr = Expr.BinaryOp(None, "Concat", [hi_value.expr, lo_value.expr], False)
                    return True, result_expr
            return False, None

        new_expr = self.state.load_register(expr)
        if new_expr is not None:
            # check if this new_expr uses any expression that has been overwritten
            replaced = False
            outdated = False
            all_subexprs = list(new_expr.all_exprs())
            for _, detail in new_expr.offset_and_details.items():
                if detail.expr is None:
                    break
                if self.is_using_outdated_def(detail.expr, detail.def_at, avoid=expr):
                    outdated = True
                    break

            if all_subexprs and None not in all_subexprs and not outdated:
                if len(all_subexprs) == 1:
                    # trivial case
                    subexpr = all_subexprs[0]
                    if subexpr.size == expr.size:
                        replaced = True
                        l.debug("Add a replacement: %s with %s", expr, subexpr)
                        self.state.add_replacement(self._codeloc(), expr, subexpr)
                else:
                    is_concatenation, result_expr = _test_concatenation(new_expr)
                    if is_concatenation:
                        replaced = True
                        l.debug("Add a replacement: %s with %s", expr, result_expr)
                        self.state.add_replacement(self._codeloc(), expr, result_expr)

            if not replaced:
                l.debug("Add a replacement: %s with TOP", expr)
                self.state.add_replacement(self._codeloc(), expr, self.state.top(expr.bits))
            else:
                return new_expr

        return PropValue.from_value_and_details(self.state.top(expr.bits), expr.size, expr, self._codeloc())

    def _ail_handle_Load(self, expr: Expr.Load) -> Optional[PropValue]:

        self.state: 'PropagatorAILState'

        addr = self._expr(expr.addr)

        addr_expr = addr.one_expr
        var_defat = None

        if addr_expr is not None:
            if isinstance(addr_expr, Expr.StackBaseOffset) and not isinstance(expr.addr, Expr.StackBaseOffset):
                l.debug("Add a replacement: %s with %s", expr.addr, addr_expr)
                self.state.add_replacement(self._codeloc(), expr.addr, addr_expr)

            sp_offset = self.extract_offset_to_sp(addr_expr)
            if sp_offset is not None:
                # Stack variable.
                var = self.state.load_stack_variable(sp_offset, expr.size, endness=expr.endness)
                if var is not None:
                    var_defat = var.one_defat
                    # We do not add replacements here since in AIL function and block simplifiers we explicitly forbid
                    # replacing stack variables, unless this is the parameter of a call (indicated by expr.func_arg is
                    # True).
                    if getattr(expr, "func_arg", False) is True \
                            or (self.state._gp is not None
                                and not self.state.is_top(var.value)
                                and var.value.concrete
                                and var.value._model_concrete.value == self.state._gp):
                        if var.one_expr is not None:
                            if not self.is_using_outdated_def(var.one_expr, var.one_defat, avoid=expr.addr):
                                l.debug("Add a replacement: %s with %s", expr, var.one_expr)
                                self.state.add_replacement(self._codeloc(), expr, var.one_expr)
                        else:
                            # there isn't a single expression to replace with. remove the old replacement for this
                            # expression if available.
                            self.state.add_replacement(self._codeloc(), expr, self.state.top(expr.bits))
                    if not self.state.is_top(var.value):
                        return var

        if addr_expr is not None and addr_expr is not expr.addr:
            new_expr = Expr.Load(expr.idx, addr_expr, expr.size, expr.endness, **expr.tags)
        else:
            new_expr = expr
        prop_value = PropValue.from_value_and_details(
            self.state.top(expr.size * self.arch.byte_width), expr.size, new_expr,
            self._codeloc() if var_defat is None else var_defat
        )
        return prop_value

    def _ail_handle_Convert(self, expr: Expr.Convert) -> PropValue:
        o_value = self._expr(expr.operand)

        if o_value is None or self.state.is_top(o_value.value):
            new_value = self.state.top(expr.to_bits)
        else:
            if expr.from_bits < expr.to_bits:
                if expr.is_signed:
                    new_value = claripy.SignExt(expr.to_bits - expr.from_bits, o_value.value)
                else:
                    new_value = claripy.ZeroExt(expr.to_bits - expr.from_bits, o_value.value)
            elif expr.from_bits > expr.to_bits:
                new_value = claripy.Extract(expr.to_bits - 1, 0, o_value.value)
            else:
                new_value = o_value.value

        o_expr = o_value.one_expr
        o_defat = o_value.one_defat
        if o_expr is not None:
            # easy cases
            if type(o_expr) is Expr.Convert:
                if expr.from_bits == o_expr.to_bits and expr.to_bits == o_expr.from_bits:
                    # eliminate the redundant Convert
                    new_expr = o_expr.operand
                else:
                    new_expr = Expr.Convert(expr.idx, o_expr.from_bits, expr.to_bits, expr.is_signed, o_expr.operand)
            elif type(o_expr) is Expr.Const:
                # do the conversion right away
                value = o_expr.value
                mask = (2 ** expr.to_bits) - 1
                value &= mask
                new_expr = Expr.Const(expr.idx, o_expr.variable, value, expr.to_bits)
            else:
                new_expr = Expr.Convert(expr.idx, expr.from_bits, expr.to_bits, expr.is_signed, o_expr, **expr.tags)

            if isinstance(new_expr, Expr.Convert) and not new_expr.is_signed \
                    and new_expr.to_bits > new_expr.from_bits and new_expr.from_bits % self.arch.byte_width == 0:
                # special handling for zero-extension: it simplifies the code if we explicitly model zeros
                new_size = new_expr.from_bits // self.arch.byte_width
                offset_and_details = {
                    0: Detail(new_size, new_expr.operand, o_defat),
                    new_size: Detail(
                        new_expr.size - new_size,
                        Expr.Const(expr.idx, None, 0, new_expr.to_bits - new_expr.from_bits),
                        self._codeloc()),
                }
            else:
                offset_and_details = {0: Detail(expr.size, new_expr, self._codeloc())}

            return PropValue(new_value, offset_and_details=offset_and_details)

        elif o_value.offset_and_details:
            # hard cases... we will keep certain labels and eliminate other labels
            start_offset = 0
            end_offset = expr.to_bits // self.arch.byte_width  # end_offset is exclusive
            offset_and_details = {}
            max_offset = max(o_value.offset_and_details.keys())
            for offset_, detail_ in o_value.offset_and_details.items():
                if offset_ < start_offset < offset_ + detail_.size:
                    # we start here
                    off = 0
                    siz = min(end_offset, offset_ + detail_.size) - start_offset
                    expr_ = PropValue.extract_ail_expression(
                        (start_offset - offset_) * self.arch.byte_width,
                        siz * self.arch.byte_width,
                        detail_.expr
                    )
                    offset_and_details[off] = Detail(siz, expr_, detail_.def_at)
                elif offset_ >= start_offset and offset_ + detail_.size <= end_offset:
                    # we include the whole thing
                    off = offset_ - start_offset
                    siz = detail_.size
                    if off == max_offset and off + siz < end_offset:
                        # extend the expr
                        expr_ = PropValue.extend_ail_expression(
                            (end_offset - (off + siz)) * self.arch.byte_width,
                            detail_.expr
                        )
                        siz = end_offset - off
                    else:
                        expr_ = detail_.expr
                    offset_and_details[off] = Detail(siz, expr_, detail_.def_at)
                elif offset_ < end_offset <= offset_ + detail_.size:
                    # we include all the way until end_offset
                    if offset_ < start_offset:
                        off = 0
                        siz = end_offset - start_offset
                    else:
                        off = offset_ - start_offset
                        siz = end_offset - offset_
                    expr_ = PropValue.extract_ail_expression(0, siz * self.arch.byte_width, detail_.expr)
                    offset_and_details[off] = Detail(siz, expr_, detail_.def_at)

            return PropValue(
                new_value,
                offset_and_details=offset_and_details
            )
        else:
            # it's empty... no expression is available for whatever reason
            return PropValue.from_value_and_details(new_value, expr.size, expr, self._codeloc())

    def _ail_handle_Const(self, expr: Expr.Const) -> PropValue:
        if isinstance(expr.value, float):
            v = claripy.FPV(expr.value, claripy.FSORT_DOUBLE if expr.bits == 64 else claripy.FSORT_FLOAT)
        else:
            v = claripy.BVV(expr.value, expr.bits)
        return PropValue.from_value_and_details(v, expr.size, expr, self._codeloc())

    def _ail_handle_DirtyExpression(self, expr: Expr.DirtyExpression) -> Optional[PropValue]:  # pylint:disable=no-self-use

        if isinstance(expr.dirty_expr, Expr.VEXCCallExpression):
            for operand in expr.dirty_expr.operands:
                _ = self._expr(operand)

        return PropValue.from_value_and_details(
            self.state.top(expr.bits), expr.size, expr, self._codeloc()
        )

    def _ail_handle_ITE(self, expr: Expr.ITE) -> Optional[PropValue]:
        # pylint:disable=unused-variable
        cond = self._expr(expr.cond)
        iftrue = self._expr(expr.iftrue)
        iffalse = self._expr(expr.iffalse)

        return PropValue.from_value_and_details(
            self.state.top(expr.bits),
            expr.size, expr, self._codeloc()
        )

    def _ail_handle_Reinterpret(self, expr: Expr.Reinterpret) -> Optional[PropValue]:
        arg = self._expr(expr.operand)

        if self.state.is_top(arg.value):
            one_expr = arg.one_expr
            if one_expr is not None:
                expr = Expr.Reinterpret(expr.idx, expr.from_bits, expr.from_type, expr.to_bits, expr.to_type, one_expr,
                                        **expr.tags)

        return PropValue.from_value_and_details(
            arg.value,
            expr.size, expr, self._codeloc()
        )

    def _ail_handle_CallExpr(self, expr_stmt: Stmt.Call) -> Optional[PropValue]:
        if isinstance(expr_stmt.target, Expr.Expression):
            _ = self._expr(expr_stmt.target)

        if expr_stmt.args:
            for arg in expr_stmt.args:
                _ = self._expr(arg)

        # ignore ret_expr
        return PropValue.from_value_and_details(
            self.state.top(expr_stmt.bits),
            expr_stmt.size, expr_stmt, self._codeloc()
        )

    def _ail_handle_Not(self, expr):
        o_value = self._expr(expr.operand)

        value = self.state.top(expr.bits)
        if o_value is None:
            new_expr = expr
        else:
            o_expr = o_value.one_expr
            new_expr = Expr.UnaryOp(expr.idx,
                                    'Not',
                                     o_expr if o_expr is not None else expr.operands[0],
                                     **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_Neg(self, expr):
        o_value = self._expr(expr.operand)

        value = self.state.top(expr.bits)
        if o_value is None:
            new_expr = expr
        else:
            o_expr = o_value.one_expr
            new_expr = Expr.UnaryOp(expr.idx,
                                    'Neg',
                                     o_expr if o_expr is not None else expr.operands[0],
                                     **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_Cmp(self, expr: Expr.BinaryOp) -> PropValue:
        operand_0_value = self._expr(expr.operands[0])
        operand_1_value = self._expr(expr.operands[1])

        if operand_0_value is not None and operand_1_value is not None:
            operand_0_oneexpr = operand_0_value.one_expr
            operand_1_oneexpr = operand_1_value.one_expr
            if operand_0_oneexpr is expr.operands[0] and operand_1_oneexpr is expr.operands[1]:
                # nothing changed
                return PropValue.from_value_and_details(self.state.top(expr.bits), expr.size, expr, self._codeloc())
            else:
                operand_0 = operand_0_oneexpr if operand_0_oneexpr is not None else expr.operands[0]
                operand_1 = operand_1_oneexpr if operand_1_oneexpr is not None else expr.operands[1]

            new_expr = Expr.BinaryOp(expr.idx, expr.op, [operand_0, operand_1], expr.signed, **expr.tags)
        else:
            new_expr = expr

        return PropValue.from_value_and_details(
            self.state.top(expr.bits),
            expr.size, new_expr, self._codeloc()
        )

    _ail_handle_CmpF = _ail_handle_Cmp
    _ail_handle_CmpLE = _ail_handle_Cmp
    _ail_handle_CmpLEs = _ail_handle_Cmp
    _ail_handle_CmpLT = _ail_handle_Cmp
    _ail_handle_CmpLTs = _ail_handle_Cmp
    _ail_handle_CmpGE = _ail_handle_Cmp
    _ail_handle_CmpGEs = _ail_handle_Cmp
    _ail_handle_CmpGT = _ail_handle_Cmp
    _ail_handle_CmpGTs = _ail_handle_Cmp
    _ail_handle_CmpEQ = _ail_handle_Cmp
    _ail_handle_CmpNE = _ail_handle_Cmp

    def _ail_handle_Add(self, expr: Expr.BinaryOp) -> PropValue:
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        if o0_value is None or o1_value is None:
            new_expr = expr
            value = self.state.top(expr.bits)
        else:
            if o0_value.value.concrete and o1_value.value.concrete:
                value = o0_value.value + o1_value.value
            else:
                value = self.state.top(expr.bits)

            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            if isinstance(o0_expr, Expr.BasePointerOffset) and isinstance(o1_expr, Expr.Const):
                new_expr = o0_value.one_expr.copy()
                new_expr.offset += o1_expr.value
            else:
                new_expr = Expr.BinaryOp(expr.idx,
                                         'Add',
                                         [o0_expr if o0_expr is not None else expr.operands[0],
                                          o1_expr if o1_expr is not None else expr.operands[1],],
                                         expr.signed,
                                         floating_point=expr.floating_point,
                                         rounding_mode=expr.rounding_mode,
                                         **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_Sub(self, expr: Expr.BinaryOp) -> PropValue:
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        if o0_value is None or o1_value is None:
            new_expr = expr
            value = self.state.top(expr.bits)
        else:
            if o0_value.value.concrete and o1_value.value.concrete:
                value = o0_value.value - o1_value.value
            else:
                value = self.state.top(expr.bits)

            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            if isinstance(o0_expr, Expr.BasePointerOffset) and isinstance(o1_expr, Expr.Const):
                new_expr = o0_value.one_expr.copy()
                new_expr.offset -= o1_expr.value
            else:
                new_expr = Expr.BinaryOp(expr.idx,
                                         'Sub',
                                         [o0_expr if o0_expr is not None else expr.operands[0],
                                          o1_expr if o1_expr is not None else expr.operands[1],],
                                         expr.signed,
                                         floating_point=expr.floating_point,
                                         rounding_mode=expr.rounding_mode,
                                         **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_StackBaseOffset(self, expr: Expr.StackBaseOffset) -> PropValue:  # pylint:disable=no-self-use
        return PropValue.from_value_and_details(self.state.top(expr.bits), expr.size, expr, self._codeloc())

    def _ail_handle_And(self, expr: Expr.BinaryOp):

        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        value = self.state.top(expr.bits)
        if o0_value is None or o1_value is None:
            new_expr = expr
        else:
            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr

            # Special logic for stack pointer alignment
            sp_offset = self.extract_offset_to_sp(o0_value.value)
            if sp_offset is not None and type(o1_expr) is Expr.Const and is_alignment_mask(o1_expr.value):
                value = o0_value.value
                new_expr = o0_expr
            elif isinstance(o0_expr, Expr.StackBaseOffset) and type(o1_expr) is Expr.Const \
                    and is_alignment_mask(o1_expr.value):
                value = o0_value.value
                new_expr = o0_expr
            else:
                value = self.state.top(expr.bits)
                new_expr = Expr.BinaryOp(expr.idx,
                                         'And',
                                         [o0_expr if o0_expr is not None else expr.operands[0],
                                          o1_expr if o1_expr is not None else expr.operands[1],],
                                         expr.signed,
                                         floating_point=expr.floating_point,
                                         rounding_mode=expr.rounding_mode,
                                         **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_Or(self, expr: Expr.BinaryOp):
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        value = self.state.top(expr.bits)
        if o0_value is None or o1_value is None:
            new_expr = expr
        else:
            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            new_expr = Expr.BinaryOp(expr.idx,
                                     'Or',
                                     [o0_expr if o0_expr is not None else expr.operands[0],
                                      o1_expr if o1_expr is not None else expr.operands[1],],
                                     expr.signed,
                                     floating_point=expr.floating_point,
                                     rounding_mode=expr.rounding_mode,
                                     **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_Xor(self, expr: Expr.BinaryOp):
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        value = self.state.top(expr.bits)
        if o0_value is None or o1_value is None:
            new_expr = expr
        else:
            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            new_expr = Expr.BinaryOp(expr.idx,
                                     'Xor',
                                     [o0_expr if o0_expr is not None else expr.operands[0],
                                      o1_expr if o1_expr is not None else expr.operands[1],],
                                     expr.signed,
                                     floating_point=expr.floating_point,
                                     rounding_mode=expr.rounding_mode,
                                     **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_Shl(self, expr: Expr.BinaryOp):
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        value = self.state.top(expr.bits)
        if o0_value is None or o1_value is None:
            new_expr = expr
        else:
            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            new_expr = Expr.BinaryOp(expr.idx,
                                     'Shl',
                                     [o0_expr if o0_expr is not None else expr.operands[0],
                                      o1_expr if o1_expr is not None else expr.operands[1], ],
                                     expr.signed,
                                     floating_point=expr.floating_point,
                                     rounding_mode=expr.rounding_mode,
                                     **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_Shr(self, expr: Expr.BinaryOp):
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        value = self.state.top(expr.bits)
        if o0_value is None or o1_value is None:
            new_expr = expr
        else:
            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            new_expr = Expr.BinaryOp(expr.idx,
                                     'Shr',
                                     [o0_expr if o0_expr is not None else expr.operands[0],
                                      o1_expr if o1_expr is not None else expr.operands[1],],
                                     expr.signed,
                                     floating_point=expr.floating_point,
                                     rounding_mode=expr.rounding_mode,
                                     **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_Sar(self, expr: Expr.BinaryOp):
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        value = self.state.top(expr.bits)
        if o0_value is None or o1_value is None:
            new_expr = expr
        else:
            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            new_expr = Expr.BinaryOp(expr.idx,
                                     'Sar',
                                     [o0_expr if o0_expr is not None else expr.operands[0],
                                      o1_expr if o1_expr is not None else expr.operands[1],],
                                     expr.signed,
                                     floating_point=expr.floating_point,
                                     rounding_mode=expr.rounding_mode,
                                     **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_Mul(self, expr):
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        value = self.state.top(expr.bits)
        if o0_value is None or o1_value is None:
            new_expr = expr
        else:
            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            new_expr = Expr.BinaryOp(expr.idx,
                                     'Mul',
                                     [o0_expr if o0_expr is not None else expr.operands[0],
                                      o1_expr if o1_expr is not None else expr.operands[1],],
                                     expr.signed,
                                     floating_point=expr.floating_point,
                                     rounding_mode=expr.rounding_mode,
                                     **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_Mull(self, expr):
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        value = self.state.top(expr.bits)
        if o0_value is None or o1_value is None:
            new_expr = expr
        else:
            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            new_expr = Expr.BinaryOp(expr.idx,
                                     'Mull',
                                     [o0_expr if o0_expr is not None else expr.operands[0],
                                      o1_expr if o1_expr is not None else expr.operands[1],],
                                     expr.signed,
                                     floating_point=expr.floating_point,
                                     rounding_mode=expr.rounding_mode,
                                     **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_Div(self, expr):
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        value = self.state.top(expr.bits)
        if o0_value is None or o1_value is None:
            new_expr = expr
        else:
            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            new_expr = Expr.BinaryOp(expr.idx,
                                     'Div',
                                     [o0_expr if o0_expr is not None else expr.operands[0],
                                      o1_expr if o1_expr is not None else expr.operands[1],],
                                     expr.signed,
                                     floating_point=expr.floating_point,
                                     rounding_mode=expr.rounding_mode,
                                     **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_DivMod(self, expr):
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        value = self.state.top(expr.bits)
        if o0_value is None or o1_value is None:
            new_expr = expr
        else:
            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            new_expr = Expr.BinaryOp(expr.idx,
                                     'DivMod',
                                     [o0_expr if o0_expr is not None else expr.operands[0],
                                      o1_expr if o1_expr is not None else expr.operands[1],],
                                     expr.signed,
                                     floating_point=expr.floating_point,
                                     rounding_mode=expr.rounding_mode,
                                     **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    def _ail_handle_TernaryOp(self, expr: Expr.TernaryOp):
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])
        o2_value = self._expr(expr.operands[2])

        if o0_value is None or o1_value is None or o2_value is None:
            new_expr = expr
        else:
            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            o2_expr = o2_value.one_expr
            new_expr = Expr.TernaryOp(
                expr.idx,
                expr.op,
                [o0_expr if o0_expr is not None else expr.operands[0],
                 o1_expr if o1_expr is not None else expr.operands[1],
                 o2_expr if o2_expr is not None else expr.operands[2]],
                bits=expr.bits,
                **expr.tags
            )

        return PropValue.from_value_and_details(self.state.top(expr.bits), expr.size, new_expr, self._codeloc())

    def _ail_handle_Concat(self, expr):
        o0_value = self._expr(expr.operands[0])
        o1_value = self._expr(expr.operands[1])

        value = self.state.top(expr.bits)
        if o0_value is None or o1_value is None:
            new_expr = expr
        else:
            o0_expr = o0_value.one_expr
            o1_expr = o1_value.one_expr
            new_expr = Expr.BinaryOp(expr.idx,
                                     'Concat',
                                     [o0_expr if o0_expr is not None else expr.operands[0],
                                      o1_expr if o1_expr is not None else expr.operands[1], ],
                                     expr.signed,
                                     **expr.tags)
        return PropValue.from_value_and_details(value, expr.size, new_expr, self._codeloc())

    #
    # Util methods
    #

    def is_using_outdated_def(self, expr: Expr.Expression, expr_defat: 'CodeLocation',
                              avoid: Optional[Expr.Expression]=None) -> bool:

        from .outdated_definition_walker import OutdatedDefinitionWalker  # pylint:disable=import-outside-toplevel

        walker = OutdatedDefinitionWalker(expr, expr_defat, self.state,
                                          avoid=avoid,
                                          extract_offset_to_sp=self.extract_offset_to_sp)
        walker.walk_expression(expr)
        return walker.out_dated



from ailment.expression import BinaryOp, Const

from .base import PeepholeOptimizationExprBase


class ADivConstAddAMulNDivConst(PeepholeOptimizationExprBase):
    __slots__ = ()

    NAME = "a / N0 + (a * N1) / N0 => a * (N1 + 1) / N0"
    expr_classes = (BinaryOp, )

    def optimize(self, expr: BinaryOp):

        if expr.op == "Add" and len(expr.operands) == 2:
            op0, op1 = expr.operands
            if isinstance(op0, BinaryOp) and op0.op == "Div" and isinstance(op0.operands[1], Const):
                if isinstance(op1, BinaryOp) and op1.op == "Div" and isinstance(op1.operands[1], Const):
                    if isinstance(op1.operands[0], BinaryOp) and op1.operands[0].op == "Mul" and \
                            isinstance(op1.operands[0].operands[1], Const):
                        a0 = op0.operands[0]
                        a1 = op1.operands[0].operands[0]
                        if a0.likes(a1):
                            N0 = op0.operands[1]
                            N1: int = op1.operands[0].operands[1].value
                            if N0.value == op1.operands[1].value:
                                mul = BinaryOp(op0.idx, "Mul",
                                               [a0,
                                                Const(None, None, N1 + 1, expr.bits, **expr.operands[0].operands[1].tags)
                                                ],
                                               False,
                                               **op0.tags)
                                div = BinaryOp(expr.idx, "Div",
                                               [mul,
                                                N0],
                                               False,
                                               **expr.tags)
                                return div

        return None




import logging
import re
from collections import defaultdict

from . import Analysis, CFGEmulated, DDG

from ..knowledge_base import KnowledgeBase
from .. import SIM_PROCEDURES
from ..codenode import HookNode
from ..sim_variable import SimConstantVariable, SimRegisterVariable, SimMemoryVariable, SimStackVariable

l = logging.getLogger(name=__name__)


class ConstantPropagation(object):
    def __init__(self, constant, constant_assignment_loc, constant_consuming_loc):
        self.constant = constant
        self.constant_assignment_loc = constant_assignment_loc
        self.constant_consuming_loc = constant_consuming_loc

    def __repr__(self):
        s = "<Constant %#x propagates from %#x to %#x>" % (
            self.constant,
            self.constant_assignment_loc.ins_addr,
            self.constant_consuming_loc.ins_addr
        )

        return s


class RedundantStackVariable(object):
    def __init__(self, argument, stack_variable, stack_variable_consuming_locs):
        self.argument = argument
        self.stack_variable = stack_variable
        self.stack_variable_consuming_locs = stack_variable_consuming_locs
        self.argument_register_as_retval = False

    def __repr__(self):
        s = "<StackVar %s for %s at %d locations%s>" % (
            self.stack_variable,
            self.argument,
            len(self.stack_variable_consuming_locs),
            " - retval" if self.argument_register_as_retval else "",
        )

        return s


class RegisterReallocation(object):
    def __init__(self, stack_variable, register_variable, stack_variable_sources, stack_variable_consumers,
                 prologue_addr, prologue_size, epilogue_addr, epilogue_size):
        """
        Constructor.

        :param SimStackVariable stack_variable:
        :param SimRegisterVariable register_variable:
        :param list stack_variable_sources:
        :param list stack_variable_consumers:
        :param int prologue_addr:
        :param int prologue_size:
        :param int epilogue_addr:
        :param int epilogue_size:
        """

        self.stack_variable = stack_variable
        self.register_variable = register_variable
        self.stack_variable_sources = stack_variable_sources
        self.stack_variable_consumers = stack_variable_consumers

        self.prologue_addr = prologue_addr
        self.prologue_size = prologue_size
        self.epilogue_addr = epilogue_addr
        self.epilogue_size = epilogue_size

    def __repr__(self):
        s = "<RegisterReallocation %s for %s with %d sources and %d consumers>" % (
            self.register_variable,
            self.stack_variable,
            len(self.stack_variable_sources),
            len(self.stack_variable_consumers),
        )
        return s


class DeadAssignment(object):
    def __init__(self, pv):
        """
        Constructor.

        :param angr.analyses.ddg.ProgramVariable pv: The assignment to remove.
        """

        self.pv = pv

    def __repr__(self):
        s = "<DeadAssignmentElimination %s>" % self.pv
        return s


class BinaryOptimizer(Analysis):
    """
    This is a collection of binary optimization techniques we used in Mechanical Phish during the finals of Cyber Grand
    Challange. It focuses on dealing with some serious speed-impacting code constructs, and *sort of* worked on *some*
    CGC binaries compiled with O0. Use this analysis as a reference of how to use data dependency graph and such.

    There is no guarantee that BinaryOptimizer will ever work on non-CGC binaries. Feel free to give us PR or MR, but
    please *do not* ask for support of non-CGC binaries.
    """

    BLOCKS_THRESHOLD = 500  # do not optimize a function if it has more than this number of blocks

    def __init__(self, cfg, techniques):

        self.cfg = cfg

        if techniques is None:
            raise Exception('At least one optimization technique must be specified.')

        supported_techniques = {
            'constant_propagation',
            'redundant_stack_variable_removal',
            'register_reallocation',
            'dead_assignment_elimination',
        }

        if techniques - supported_techniques:
            raise Exception('At least one optimization technique specified is not supported.')

        self._techniques = techniques.copy()

        self.constant_propagations = [ ]
        self.redundant_stack_variables = [ ]
        self.register_reallocations = [ ]
        self.dead_assignments = [ ]

        self.optimize()

    def optimize(self):
        for f in self.kb.functions.values():  # type: angr.knowledge.Function
            # if there are unresolved targets in this function, we do not try to optimize it
            unresolvable_targets = (SIM_PROCEDURES['stubs']['UnresolvableJumpTarget'],
                                    SIM_PROCEDURES['stubs']['UnresolvableCallTarget'])
            if any([ n.sim_procedure in unresolvable_targets for n in f.graph.nodes()
                     if isinstance(n, HookNode) ]):
                continue

            if len(f.block_addrs_set) > self.BLOCKS_THRESHOLD:
                continue

            self._optimize_function(f)

    def _optimize_function(self, function):
        """

        :param angr.knowledge.Function function:
        :return:
        """

        #if function.addr != 0x8048250:
        #    return

        func_kb = KnowledgeBase(self.project)

        # switch to non-optimized IR, since optimized IR will optimize away register reads/writes
        # for example,
        # .text:08048285 add     eax, [ebp+var_8]
        # .text:08048288 mov     [ebp+var_C], eax
        # becomes
        #    06 | ------ IMark(0x8048285, 3, 0) ------
        #    07 | t25 = Add32(t24,0xfffffff8)
        #    08 | t5 = LDle:I32(t25)
        #    09 | t4 = Add32(t2,t5)
        #    10 | PUT(eip) = 0x08048288
        #    11 | ------ IMark(0x8048288, 3, 0) ------
        #    12 | t27 = Add32(t24,0xfffffff4)
        #    13 | STle(t27) = t4
        #    14 | PUT(eip) = 0x0804828b
        # there is no write to or read from eax

        cfg = self.project.analyses[CFGEmulated].prep(kb=func_kb)(
                                                call_depth=1,
                                                base_graph=function.graph,
                                                keep_state=True,
                                                starts=(function.addr,),
                                                iropt_level=0,
                                                )

        ddg = self.project.analyses[DDG].prep(kb=func_kb)(cfg=cfg)

        if 'constant_propagation' in self._techniques:
            self._constant_propagation(function, ddg.simplified_data_graph)
        if 'redundant_stack_variable_removal' in self._techniques:
            self._redundant_stack_variable_removal(function, ddg.simplified_data_graph)
        if 'register_reallocation' in self._techniques:
            self._register_reallocation(function, ddg.simplified_data_graph)
        if 'dead_assignment_elimination' in self._techniques:
            self._dead_assignment_elimination(function, ddg.simplified_data_graph)

    def _constant_propagation(self, function, data_graph):  #pylint:disable=unused-argument
        """

        :param function:
        :param networkx.MultiDiGraph data_graph:
        :return:
        """

        # find all edge sequences that looks like const->reg->memory

        for n0 in data_graph.nodes():
            if not isinstance(n0.variable, SimConstantVariable):
                continue

            n1s = list(data_graph.successors(n0))
            if len(n1s) != 1:
                continue
            n1 = n1s[0]

            if not isinstance(n1.variable, SimRegisterVariable):
                continue
            if len(list(data_graph.predecessors(n1))) != 1:
                continue

            n2s = list(data_graph.successors(n1))
            if len(n2s) != 1:
                continue
            n2 = n2s[0]

            if not isinstance(n2.variable, SimMemoryVariable):
                continue
            n2_inedges = data_graph.in_edges(n2, data=True)
            if len([ 0 for _, _, data in n2_inedges if 'type' in data and data['type'] == 'mem_data' ]) != 1:
                continue

            cp = ConstantPropagation(n0.variable.value, n0.location, n2.location)
            self.constant_propagations.append(cp)

            # print n0, n1, n2

    def _redundant_stack_variable_removal(self, function, data_graph):
        """
        If an argument passed from the stack (i.e. dword ptr [ebp+4h]) is saved to a local variable on the stack at the
        beginning of the function, and this local variable was never modified anywhere in this function, and no pointer
        of any stack variable is saved in any register, then we can replace all references to this local variable to
        that argument instead.

        :param function:
        :param networkx.MultiDiGraph data_graph:
        :return:
        """

        # check if there is any stack pointer being stored into any register other than esp
        # basically check all consumers of stack pointers
        stack_ptrs = [ ]
        sp_offset = self.project.arch.registers['esp'][0]
        bp_offset = self.project.arch.registers['ebp'][0]
        for n in data_graph.nodes():
            if isinstance(n.variable, SimRegisterVariable) and n.variable.reg in (sp_offset, bp_offset):
                stack_ptrs.append(n)

        # for each stack pointer variable, make sure none of its consumers is a general purpose register
        for stack_ptr in stack_ptrs:
            out_edges = data_graph.out_edges(stack_ptr, data=True)
            for _, dst, data in out_edges:
                if 'type' in data and data['type'] == 'kill':
                    # we don't care about killing edges
                    continue
                if isinstance(dst.variable, SimRegisterVariable) and dst.variable.reg < 40 and \
                        dst.variable.reg not in (sp_offset, bp_offset):
                    # oops
                    l.debug('Function %s does not satisfy requirements of redundant stack variable removal.',
                            repr(function)
                            )
                    return

        argument_variables = [ ]

        for n in data_graph.nodes():
            if isinstance(n.variable, SimStackVariable) and n.variable.base == 'bp' and n.variable.offset >= 0:
                argument_variables.append(n)

        if not argument_variables:
            return

        #print function
        #print argument_variables

        argument_to_local = { }
        argument_register_as_retval = set()

        # for each argument, find its correspondence on the local stack frame

        for argument_variable in argument_variables:
            # is it copied to the stack?
            successors0 = list(data_graph.successors(argument_variable))

            if not successors0:
                continue

            if len(successors0) != 1:
                continue

            if isinstance(successors0[0].variable, SimRegisterVariable):
                # argument -> register -> stack
                out_edges = data_graph.out_edges(successors0[0], data=True)
                successors1 = [ s for _, s, data in out_edges if 'type' not in data or data['type'] != 'kill' ]
                if len(successors1) == 1:
                    successor1 = successors1[0]
                    if isinstance(successor1.variable, SimStackVariable):
                        if (successor1.variable.base == 'sp' and successor1.variable.offset > 0) or \
                                (successor1.variable.base == 'bp' and successor1.variable.offset < 0):
                            # yes it's copied onto the stack!
                            argument_to_local[argument_variable] = successor1

                # if the register is eax, and it's not killed later, it might be the return value of this function
                # in that case, we cannot eliminate the instruction that moves stack argument to that register
                if successors0[0].variable.reg == self.project.arch.registers['eax'][0]:
                    killers = [ s for _, s, data in out_edges if 'type' in data and data['type'] == 'kill']
                    if not killers:
                        # it might be the return value
                        argument_register_as_retval.add(argument_variable)

            else:
                # TODO:
                import ipdb; ipdb.set_trace()

        #import pprint
        #pprint.pprint(argument_to_local, width=160)

        # find local correspondence that are not modified throughout this function
        redundant_stack_variables = [ ]

        for argument, local_var in argument_to_local.items():
            # local_var cannot be killed anywhere
            out_edges = data_graph.out_edges(local_var, data=True)

            consuming_locs = [ ]

            for _, consumer, data in out_edges:
                consuming_locs.append(consumer.location)
                if 'type' in data and data['type'] == 'kill':
                    break
            else:
                # no killing edges. the value is not changed!
                rsv = RedundantStackVariable(argument, local_var, consuming_locs)
                if argument in argument_register_as_retval:
                    rsv.argument_register_as_retval = True
                redundant_stack_variables.append(rsv)

        self.redundant_stack_variables.extend(redundant_stack_variables)

    def _register_reallocation(self, function, data_graph):
        """
        Find unused registers throughout the function, and use those registers to replace stack variables.

        Only functions that satisfy the following criteria can be optimized in this way:
        - The function does not call any other function.
        - The function does not use esp to index any stack variable.
        - Prologue and epilogue of the function is identifiable.
        - At least one register is not used in the entire function.

        :param angr.knowledge.Function function:
        :param networkx.MultiDiGraph data_graph:
        :return: None
        """

        # make sure this function does not call other functions
        if function.callout_sites:
            return

        if len(function.endpoints) != 1:
            return

        # identify function prologue and epilogue
        startpoint_block = self.project.factory.block(function.startpoint.addr).capstone
        startpoint_insns = startpoint_block.insns

        # supported function prologues:
        #
        # push  ebp
        # mov   ebp, esp
        # sub   esp, [0-9a-f]+h
        #
        # push  ebp
        # mov   ebp, esp
        # push  eax

        if len(startpoint_insns) < 3:
            return

        insn0, insn1, insn2 = startpoint_insns[:3]

        if not (insn0.mnemonic == 'push' and insn0.op_str == 'ebp'):
            return
        if not (insn1.mnemonic == 'mov' and insn1.op_str == 'ebp, esp'):
            return
        if not (insn2.mnemonic == 'sub' and re.match(r"esp, [0-9a-fx]+", insn2.op_str)) and \
                not (insn2.mnemonic == 'push' and insn2.op_str == 'eax'):
            return


        endpoint_block = self.project.factory.block(function.endpoints[0].addr).capstone
        endpoint_insns = endpoint_block.insns

        # supported function epilogues:
        #
        # add   esp, [0-9a-f]+h
        # pop   ebp
        # ret

        if len(endpoint_insns) < 3:
            return

        insn3, insn4, insn5 = endpoint_insns[-3:]

        if not (insn3.mnemonic == 'add' and re.match(r"esp, [0-9a-fx]+", insn3.op_str)):
            return
        if not (insn4.mnemonic == 'pop' and insn4.op_str == 'ebp'):
            return
        if not insn5.mnemonic == 'ret':
            return

        # make sure esp is not used anywhere else - all stack variables must be indexed using ebp
        esp_offset = self.project.arch.registers['esp'][0]
        ebp_offset = self.project.arch.registers['ebp'][0]
        esp_variables = [ ]
        for n in data_graph.nodes():
            if isinstance(n.variable, SimRegisterVariable) and n.variable.reg == esp_offset:
                esp_variables.append(n)

        # find out all call instructions
        call_insns = set()
        for src, dst, data in function.transition_graph.edges(data=True):
            if 'type' in data and data['type'] == 'call':
                src_block = function._get_block(src.addr)
                call_insns.add(src_block.instruction_addrs[-1])

        # there should be six esp variables + all call sites
        # push ebp (insn0 - read, insn0 - write) ; sub esp, 0xXX (insn2) ;
        # add esp, 0xXX (insn3) ; pop ebp (insn4) ; ret (insn5)

        esp_insns = set( n.location.ins_addr for n in esp_variables )
        if esp_insns != { insn0.address, insn2.address, insn3.address, insn4.address, insn5.address } | call_insns:
            return

        prologue_addr = insn0.address
        prologue_size = insn0.size + insn1.size + insn2.size
        epilogue_addr = insn3.address
        epilogue_size = insn3.size + insn4.size + insn5.size

        # look at consumer of those esp variables. no other instruction should be consuming them
        # esp_consumer_insns = { insn0.address, insn1.address, insn2.address, insn3.address, insn4.address,
        #                        insn5.address} | esp_insns
        # for esp_variable in esp_variables:  # type: angr.analyses.ddg.ProgramVariable
        #     consumers = data_graph.successors(esp_variable)
        #     if any([ consumer.location.ins_addr not in esp_consumer_insns for consumer in consumers ]):
        #         return

        # make sure we never gets the address of those stack variables into any register
        # say, lea edx, [ebp-0x4] is forbidden
        # check all edges in data graph
        for src, dst, data in data_graph.edges(data=True):
            if isinstance(dst.variable, SimRegisterVariable) and \
                    dst.variable.reg != ebp_offset and \
                    dst.variable.reg < 40:
                #to a register other than ebp
                if isinstance(src.variable, SimRegisterVariable) and \
                        src.variable.reg == ebp_offset:
                    # from ebp
                    l.debug("Found a lea operation from ebp at %#x. Function %s cannot be optimized.",
                            dst.location.ins_addr,
                            repr(function),
                            )
                    return

        # we definitely don't want to mess with fp or sse operations
        for node in data_graph.nodes():
            if isinstance(node.variable, SimRegisterVariable) and \
                    72 <= node.variable.reg < 288:  # offset(mm0) <= node.variable.reg < offset(cs)
                l.debug('Found a float-point/SSE register access at %#x. Function %s cannot be optimized.',
                        node.location.ins_addr,
                        repr(function)
                        )
                return

        l.debug("RegisterReallocation: function %s satisfies the criteria.", repr(function))

        # nice. let's see if we can optimize this function
        # do we have free registers?

        used_general_registers = set()
        for n in data_graph.nodes():
            if isinstance(n.variable, SimRegisterVariable):
                if n.variable.reg < 40:  # this is a hardcoded limit - we only care about general registers
                    used_general_registers.add(n.variable.reg)
        registers = self.project.arch.registers
        all_general_registers = { #registers['eax'][0], registers['ecx'][0], registers['edx'][0],
                                  registers['ebx'][0], registers['edi'][0], registers['esi'][0],
                                  registers['esp'][0], registers['ebp'][0]
                                  }
        unused_general_registers = all_general_registers - used_general_registers

        if not unused_general_registers:
            l.debug("RegisterReallocation: function %s does not have any free register.", repr(function))
            return
        l.debug("RegisterReallocation: function %s has %d free register(s): %s",
                repr(function),
                len(unused_general_registers),
                ", ".join([self.project.arch.register_names[u] for u in unused_general_registers ])
                )

        # find local stack variables of size 4
        stack_variables = set()
        for n in data_graph.nodes():
            if isinstance(n.variable, SimStackVariable) and \
                    n.variable.base == 'bp' and \
                    n.variable.size == 4 and \
                    n.variable.offset < 0:
                stack_variables.add(n)

        # alright, now we need to make sure that stack variables are never accessed by indexes
        # in other words, they must be accessed directly in forms of 'dword ptr [ebp+x]'
        # it's easy to do this: we get mem_addr predecessors of each stack variable, and make sure there are only two of
        # them: one is ebp, the other one is a constant
        #
        # ah, also, since we do not want to mess with crazy fp registers, we further require none of the stack variable
        # sources and consumers is a FP register.

        filtered_stack_variables = set()
        for stack_variable in stack_variables:

            failed = False

            # check how they are accessed
            in_edges = data_graph.in_edges(stack_variable, data=True)
            for src, _, data in in_edges:
                if 'type' in data and data['type'] == 'mem_addr':
                    if isinstance(src.variable, SimRegisterVariable) and src.variable.reg == ebp_offset:
                        # ebp
                        pass
                    elif isinstance(src.variable, SimConstantVariable):
                        # the constant
                        pass
                    else:
                        # ouch
                        failed = True
                        break

                if isinstance(src.variable, SimRegisterVariable) and src.variable.reg >= 72:
                    # it comes from a FP register
                    failed = True
                    break

            if failed:
                continue

            # check consumers
            out_edges = data_graph.out_edges(stack_variable, data=True)
            for _, dst, data in out_edges:
                if 'type' in data and data['type'] == 'kill':
                    continue
                if isinstance(dst.variable, SimRegisterVariable) and dst.variable.reg >= 72:
                    # an FP register is the consumer
                    failed = True
                    break

            if failed:
                continue

            filtered_stack_variables.add(stack_variable)

        # order the stack variables by the sum of their in and out degrees.
        stack_variable_to_degree = defaultdict(int)
        stack_variable_sources = defaultdict(list)
        for sv in filtered_stack_variables:
            stack_variable_to_degree[sv.variable] += data_graph.in_degree(sv)
            stack_variable_to_degree[sv.variable] += data_graph.out_degree(sv)
            stack_variable_sources[sv.variable].append(sv)

        sorted_stack_variables = sorted(stack_variable_to_degree.keys(),
                                       key=lambda sv: stack_variable_to_degree[sv],
                                       reverse=True
                                       )

        # aha these are the ones that we can replace!
        for reg, sv in zip(unused_general_registers, sorted_stack_variables):

            non_initial_sources = [src for src in stack_variable_sources[sv] if not src.initial]

            if not non_initial_sources:
                # we failed to find any source for it, which indicates a failure in our dependence analysis
                # skip
                continue

            # get consumers
            consumers = set()
            for src in stack_variable_sources[sv]:
                out_edges = data_graph.out_edges(src, data=True)
                for _, dst, data in out_edges:
                    if 'type' not in data or data['type'] != 'kill':
                        consumers.add(dst)

            rr = RegisterReallocation(sv, SimRegisterVariable(reg, 4), non_initial_sources,
                                      list(consumers), prologue_addr, prologue_size, epilogue_addr, epilogue_size
                                      )
            self.register_reallocations.append(rr)

            l.debug("RegisterReallocation: %s will replace %s in function %s.",
                    rr.register_variable,
                    rr.stack_variable,
                    repr(function)
                    )

    def _dead_assignment_elimination(self, function, data_graph):  #pylint:disable=unused-argument
        """
        Remove assignments to registers that has no consumers, but immediately killed.

        BROKEN - DO NOT USE IT

        :param angr.knowledge.Function function:
        :param networkx.MultiDiGraph data_graph:
        :return: None
        """

        register_pvs = set()
        for node in data_graph.nodes():
            if isinstance(node.variable, SimRegisterVariable) and \
                    node.variable.reg is not None and \
                    node.variable.reg < 40:
                register_pvs.add(node)

        for reg in register_pvs:
            # does it have a consumer?
            out_edges = data_graph.out_edges(reg, data=True)
            consumers = [ ]
            killers = [ ]
            for _, _, data in out_edges:
                if 'type' in data and data['type'] == 'kill':
                    killers.append(data)
                else:
                    consumers.append(data)

            if not consumers and killers:
                # we can remove the assignment!
                da = DeadAssignment(reg)
                self.dead_assignments.append(da)

from angr.analyses import AnalysesHub
AnalysesHub.register_default('BinaryOptimizer', BinaryOptimizer)



# pylint:disable=line-too-long
import logging

from ...sim_type import SimTypeFunction,     SimTypeShort, SimTypeInt, SimTypeLong, SimTypeLongLong, SimTypeDouble, SimTypeFloat,     SimTypePointer,     SimTypeChar,     SimStruct,     SimTypeFixedSizeArray,     SimTypeBottom,     SimUnion,     SimTypeBool
from ...calling_conventions import SimCCStdcall, SimCCMicrosoftAMD64
from .. import SIM_PROCEDURES as P
from . import SimLibrary


_l = logging.getLogger(name=__name__)


lib = SimLibrary()
lib.set_default_cc('X86', SimCCStdcall)
lib.set_default_cc('AMD64', SimCCMicrosoftAMD64)
lib.set_library_names("prntvpt.dll")
prototypes = \
    {
        # 
        'PTQuerySchemaVersionSupport': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszPrinterName", "pMaxVersion"]),
        # 
        'PTOpenProvider': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszPrinterName", "dwVersion", "phProvider"]),
        # 
        'PTOpenProviderEx': SimTypeFunction([SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), offset=0), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pszPrinterName", "dwMaxVersion", "dwPrefVersion", "phProvider", "pUsedVersion"]),
        # 
        'PTCloseProvider': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider"]),
        # 
        'PTReleaseMemory': SimTypeFunction([SimTypePointer(SimTypeBottom(label="Void"), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["pBuffer"]),
        # 
        'PTGetPrintCapabilities': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeBottom(label="IStream"), SimTypeBottom(label="IStream"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider", "pPrintTicket", "pCapabilities", "pbstrErrorMessage"]),
        # 
        'PTGetPrintDeviceCapabilities': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeBottom(label="IStream"), SimTypeBottom(label="IStream"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider", "pPrintTicket", "pDeviceCapabilities", "pbstrErrorMessage"]),
        # 
        'PTGetPrintDeviceResources': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypePointer(SimTypeChar(label="Char"), offset=0), SimTypeBottom(label="IStream"), SimTypeBottom(label="IStream"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider", "pszLocaleName", "pPrintTicket", "pDeviceResources", "pbstrErrorMessage"]),
        # 
        'PTMergeAndValidatePrintTicket': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeBottom(label="IStream"), SimTypeBottom(label="IStream"), SimTypeInt(signed=False, label="EPrintTicketScope"), SimTypeBottom(label="IStream"), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider", "pBaseTicket", "pDeltaTicket", "scope", "pResultTicket", "pbstrErrorMessage"]),
        # 
        'PTConvertPrintTicketToDevMode': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeBottom(label="IStream"), SimTypeInt(signed=False, label="EDefaultDevmodeType"), SimTypeInt(signed=False, label="EPrintTicketScope"), SimTypePointer(SimTypeInt(signed=False, label="UInt32"), offset=0), SimTypePointer(SimTypePointer(SimStruct({"dmDeviceName": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32), "dmSpecVersion": SimTypeShort(signed=False, label="UInt16"), "dmDriverVersion": SimTypeShort(signed=False, label="UInt16"), "dmSize": SimTypeShort(signed=False, label="UInt16"), "dmDriverExtra": SimTypeShort(signed=False, label="UInt16"), "dmFields": SimTypeInt(signed=False, label="UInt32"), "Anonymous1": SimUnion({"Anonymous1": SimStruct({"dmOrientation": SimTypeShort(signed=True, label="Int16"), "dmPaperSize": SimTypeShort(signed=True, label="Int16"), "dmPaperLength": SimTypeShort(signed=True, label="Int16"), "dmPaperWidth": SimTypeShort(signed=True, label="Int16"), "dmScale": SimTypeShort(signed=True, label="Int16"), "dmCopies": SimTypeShort(signed=True, label="Int16"), "dmDefaultSource": SimTypeShort(signed=True, label="Int16"), "dmPrintQuality": SimTypeShort(signed=True, label="Int16")}, name="_Anonymous1_e__Struct", pack=False, align=None), "Anonymous2": SimStruct({"dmPosition": SimStruct({"x": SimTypeInt(signed=True, label="Int32"), "y": SimTypeInt(signed=True, label="Int32")}, name="POINTL", pack=False, align=None), "dmDisplayOrientation": SimTypeInt(signed=False, label="UInt32"), "dmDisplayFixedOutput": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous2_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), "dmColor": SimTypeShort(signed=True, label="Int16"), "dmDuplex": SimTypeShort(signed=True, label="Int16"), "dmYResolution": SimTypeShort(signed=True, label="Int16"), "dmTTOption": SimTypeShort(signed=True, label="Int16"), "dmCollate": SimTypeShort(signed=True, label="Int16"), "dmFormName": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32), "dmLogPixels": SimTypeShort(signed=False, label="UInt16"), "dmBitsPerPel": SimTypeInt(signed=False, label="UInt32"), "dmPelsWidth": SimTypeInt(signed=False, label="UInt32"), "dmPelsHeight": SimTypeInt(signed=False, label="UInt32"), "Anonymous2": SimUnion({"dmDisplayFlags": SimTypeInt(signed=False, label="UInt32"), "dmNup": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None"), "dmDisplayFrequency": SimTypeInt(signed=False, label="UInt32"), "dmICMMethod": SimTypeInt(signed=False, label="UInt32"), "dmICMIntent": SimTypeInt(signed=False, label="UInt32"), "dmMediaType": SimTypeInt(signed=False, label="UInt32"), "dmDitherType": SimTypeInt(signed=False, label="UInt32"), "dmReserved1": SimTypeInt(signed=False, label="UInt32"), "dmReserved2": SimTypeInt(signed=False, label="UInt32"), "dmPanningWidth": SimTypeInt(signed=False, label="UInt32"), "dmPanningHeight": SimTypeInt(signed=False, label="UInt32")}, name="DEVMODEA", pack=False, align=None), offset=0), offset=0), SimTypePointer(SimTypePointer(SimTypeChar(label="Char"), offset=0), offset=0)], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider", "pPrintTicket", "baseDevmodeType", "scope", "pcbDevmode", "ppDevmode", "pbstrErrorMessage"]),
        # 
        'PTConvertDevModeToPrintTicket': SimTypeFunction([SimTypePointer(SimTypeInt(signed=True, label="Int"), label="IntPtr", offset=0), SimTypeInt(signed=False, label="UInt32"), SimTypePointer(SimStruct({"dmDeviceName": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32), "dmSpecVersion": SimTypeShort(signed=False, label="UInt16"), "dmDriverVersion": SimTypeShort(signed=False, label="UInt16"), "dmSize": SimTypeShort(signed=False, label="UInt16"), "dmDriverExtra": SimTypeShort(signed=False, label="UInt16"), "dmFields": SimTypeInt(signed=False, label="UInt32"), "Anonymous1": SimUnion({"Anonymous1": SimStruct({"dmOrientation": SimTypeShort(signed=True, label="Int16"), "dmPaperSize": SimTypeShort(signed=True, label="Int16"), "dmPaperLength": SimTypeShort(signed=True, label="Int16"), "dmPaperWidth": SimTypeShort(signed=True, label="Int16"), "dmScale": SimTypeShort(signed=True, label="Int16"), "dmCopies": SimTypeShort(signed=True, label="Int16"), "dmDefaultSource": SimTypeShort(signed=True, label="Int16"), "dmPrintQuality": SimTypeShort(signed=True, label="Int16")}, name="_Anonymous1_e__Struct", pack=False, align=None), "Anonymous2": SimStruct({"dmPosition": SimStruct({"x": SimTypeInt(signed=True, label="Int32"), "y": SimTypeInt(signed=True, label="Int32")}, name="POINTL", pack=False, align=None), "dmDisplayOrientation": SimTypeInt(signed=False, label="UInt32"), "dmDisplayFixedOutput": SimTypeInt(signed=False, label="UInt32")}, name="_Anonymous2_e__Struct", pack=False, align=None)}, name="<anon>", label="None"), "dmColor": SimTypeShort(signed=True, label="Int16"), "dmDuplex": SimTypeShort(signed=True, label="Int16"), "dmYResolution": SimTypeShort(signed=True, label="Int16"), "dmTTOption": SimTypeShort(signed=True, label="Int16"), "dmCollate": SimTypeShort(signed=True, label="Int16"), "dmFormName": SimTypeFixedSizeArray(SimTypeChar(label="Byte"), 32), "dmLogPixels": SimTypeShort(signed=False, label="UInt16"), "dmBitsPerPel": SimTypeInt(signed=False, label="UInt32"), "dmPelsWidth": SimTypeInt(signed=False, label="UInt32"), "dmPelsHeight": SimTypeInt(signed=False, label="UInt32"), "Anonymous2": SimUnion({"dmDisplayFlags": SimTypeInt(signed=False, label="UInt32"), "dmNup": SimTypeInt(signed=False, label="UInt32")}, name="<anon>", label="None"), "dmDisplayFrequency": SimTypeInt(signed=False, label="UInt32"), "dmICMMethod": SimTypeInt(signed=False, label="UInt32"), "dmICMIntent": SimTypeInt(signed=False, label="UInt32"), "dmMediaType": SimTypeInt(signed=False, label="UInt32"), "dmDitherType": SimTypeInt(signed=False, label="UInt32"), "dmReserved1": SimTypeInt(signed=False, label="UInt32"), "dmReserved2": SimTypeInt(signed=False, label="UInt32"), "dmPanningWidth": SimTypeInt(signed=False, label="UInt32"), "dmPanningHeight": SimTypeInt(signed=False, label="UInt32")}, name="DEVMODEA", pack=False, align=None), offset=0), SimTypeInt(signed=False, label="EPrintTicketScope"), SimTypeBottom(label="IStream")], SimTypeInt(signed=True, label="Int32"), arg_names=["hProvider", "cbDevmode", "pDevmode", "scope", "pPrintTicket"]),
    }

lib.set_prototypes(prototypes)
